----------PROCEDIMENTS AUXILIARS D'ACTUALITZACI DEL MDUL ESTADSTIC----------
-- (necessari carregar-los a l'inici, doncs els procediments els necessiten) --
-------------------------------------------------------------------------------

/*****************************************************************************
** Procediment EST_ActualitzaRegalVegades (prod, rsp)                       **
******************************************************************************
** Entrada:                                                                 **
**  prod: Producte que es vol actualitzar                                   **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Actualitza el mdul estadstic. Afegeix un registre amb el producte      **
** indicat, informant d'una entrega en l'any actual. Si ja hi han entregues **
** del producte en l'any actual, les incrementa en una unitat.              **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE EST_ActualitzaRegalVegades (prod IN VARCHAR2, rsp OUT VARCHAR2)
AS

BEGIN

  rsp := 'OK';
  --Inserim una nova linia amb el producte i l'any actual. Si ja existeix,
  --la excepci s'encarrega d'actualitzar les dades.
  INSERT INTO Regal_Vegades(any_reg, producte, vegades)
    VALUES(EXTRACT(YEAR FROM SYSDATE), prod, 1);
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
    VALUES (id_log.NextVal, 'EST_ActualitzaRegalVegades', rsp, prod);
  
EXCEPTION

  --Si l'insert a la taula estadstica falla per ndex duplicat, vol dir que hi
  --ha productes entregats en l'any demanat. Incrementem el producte.
  WHEN DUP_VAL_ON_INDEX THEN
    UPDATE Regal_Vegades r
    SET r.vegades = r.vegades + 1
    WHERE r.producte = prod AND r.any_reg = EXTRACT(YEAR FROM SYSDATE);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
      VALUES (id_log.NextVal, 'EST_ActualitzaRegalVegades', rsp, prod);
    
  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
      VALUES (id_log.NextVal, 'EST_ActualitzaRegalVegades', rsp, prod);

END EST_ActualitzaRegalVegades;
/

/*****************************************************************************
** Procediment EST_ActualitzaNumRegalsAny (rsp)                             **
******************************************************************************
** Entrada:                                                                 **
**  El procediment no necessita parmetres d'entrada.                       **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Actualitza el mdul estadstic. Incrementa en una unitat el nombre de    **
** regals adquirits en l'any actual. Si no se n'han adquirit, afegeix la    **
** primera unitat.                                                          **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE EST_ActualitzaNumRegalsAny (rsp OUT VARCHAR2)
AS

BEGIN

  rsp := 'OK';
  --Inserim una nova linia amb l'any actual i un producte nou. Si ja existeix,
  --la excepci s'encarrega d'actualitzar les dades. La resta de camps son
  --inicialitzats a 0 per defecte si s necessari.
  INSERT INTO Regal_Any(any_reg, adquirits)
    VALUES(EXTRACT(YEAR FROM SYSDATE), 1);

  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1)
    VALUES (id_log.NextVal, 'EST_ActualitzaNumRegalsAny', rsp);
  
EXCEPTION

  --Si l'any ja existeix, incrementem en una unitat els regals adquirits.
  WHEN DUP_VAL_ON_INDEX THEN
  UPDATE Regal_Any r
    SET r.adquirits = r.adquirits + 1
    WHERE r.any_reg = EXTRACT(YEAR FROM SYSDATE);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1)
      VALUES (id_log.NextVal, 'EST_ActualitzaNumRegalsAny', rsp);

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1)
      VALUES (id_log.NextVal, 'EST_ActualitzaNumRegalsAny', rsp);

END EST_ActualitzaNumRegalsAny;
/

/*****************************************************************************
** Procediment EST_ActualitzaImportRegalsAny (prod, prov, rsp)              **
******************************************************************************
** Entrada:                                                                 **
**  prod (VARCHAR2): Producte que s'ha adquirit.                            **
**  prov (VARCHAR2): Proveidor del producte. Influeix en el preu del regal. **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Actualitza el mdul estadstic. Actualitza l'import en regals adquirits  **
** cada any.                                                                **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE EST_ActualitzaImportRegalsAny (prod IN VARCHAR2, prov IN VARCHAR2,
  rsp OUT VARCHAR2)
AS

  preu Regal.preu%TYPE;  --Import del producte. S'hi afegiran els impostos.
  impost Producte.impost%TYPE;  --Impost associal al producte.

BEGIN

  rsp := 'OK';
  
  --Carreguem el preu i els impostos del producte entregat.
  SELECT p.preu, p.impost INTO preu, impost
  FROM Producte p
  WHERE p.nom = prod AND p.proveidor = prov;
  
  --Calculem el preu final.
  preu := preu + (preu * impost / 100);
  --Inserim una nova linia amb l'any actual i l'import nou. Si ja existeix,
  --la excepci s'encarrega d'actualitzar les dades. La resta de camps son
  --inicialitzats a 0 per defecte si s necessari.
  INSERT INTO Regal_Any(any_reg, imp_acumulat)
    VALUES(EXTRACT(YEAR FROM SYSDATE), preu);

  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
    VALUES (id_log.NextVal, 'EST_ActualitzaImportRegalsAny', rsp, prod, prov);
  
EXCEPTION

  --Si l'any ja existeix, incrementem l'import acumulat.
  WHEN DUP_VAL_ON_INDEX THEN
  UPDATE Regal_Any r
    SET r.imp_acumulat = r.imp_acumulat + preu
    WHERE r.any_reg = EXTRACT(YEAR FROM SYSDATE);
    
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
      VALUES (id_log.NextVal, 'EST_ActualitzaImportRegalsAny', rsp, prod, prov);

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);

    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
      VALUES (id_log.NextVal, 'EST_ActualitzaImportRegalsAny', rsp, prod, prov);

END EST_ActualitzaImportRegalsAny;
/

/*****************************************************************************
** Procediment EST_ActualitzaAconseguitsAny (aconseguits, fallits, rsp)     **
******************************************************************************
** Entrada:                                                                 **
**  aconseguits (INT): Nombre de regals aconseguits pel darrer grup.        **
**  fallits (INT): Nombre de regals que han quedat en NO_SUFICIENTS_FONS    **
**   pel darrer grup de regals.                                             **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Actualitza el mdul estadstic. Mant actualitzat el percentatge de      **
** regals aconseguits respecte dels que no s'han aconseguit.                **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE EST_ActualitzaAconseguitsAny (aconseguits IN INT, fallits IN INT,
  rsp OUT VARCHAR2)
AS

  mitja Regal_any.perc_aconseguits%TYPE;
  
  SENSE_REGALS EXCEPTION;
  PRAGMA EXCEPTION_INIT(SENSE_REGALS, -1476);

BEGIN

  rsp := 'OK';
  --Calculem la mitja d'aconseguits del grup que s'acaba de tancar.
  mitja := (aconseguits * 100) / (aconseguits + fallits);
  
  --Inserim una nova linia amb l'any actual i la mitja del grup. Si ja existeix,
  --la excepci s'encarrega d'actualitzar les dades. La resta de camps son
  --inicialitzats a 0 per defecte si s necessari.
  INSERT INTO Regal_Any(any_reg, aux_aconseguits, aux_fallits, perc_aconseguits)
    VALUES(EXTRACT(YEAR FROM SYSDATE), aconseguits, fallits, mitja);

  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
    VALUES (id_log.NextVal, 'EST_ActualitzaAconseguitsAny', rsp, aconseguits,
      fallits);

EXCEPTION

  --Si l'any ja existeix, ja hi han regals aconseguits i fallits. Actualitzem
  --les dades amb les noves, i recalculem la nova mitja.
  WHEN DUP_VAL_ON_INDEX THEN
  UPDATE Regal_Any r
    SET r.aux_aconseguits = r.aux_aconseguits + aconseguits
    WHERE r.any_reg = EXTRACT(YEAR FROM SYSDATE);
  UPDATE Regal_Any r
    SET r.aux_fallits = r.aux_fallits + fallits
    WHERE r.any_reg = EXTRACT(YEAR FROM SYSDATE);
  UPDATE Regal_Any r
    SET r.perc_aconseguits = (r.aux_aconseguits * 100) /
      (r.aux_aconseguits + r.aux_fallits)
    WHERE r.any_reg = EXTRACT(YEAR FROM SYSDATE);
    
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
    VALUES (id_log.NextVal, 'EST_ActualitzaAconseguitsAny', rsp, aconseguits,
      fallits);

  --Si el grup no tenia regals, evitem la divisi entre zero i no fem res.
  WHEN SENSE_REGALS THEN
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
      VALUES (id_log.NextVal, 'EST_ActualitzaAconseguitsAny', rsp, aconseguits,
        fallits);

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
      VALUES (id_log.NextVal, 'EST_ActualitzaAconseguitsAny', rsp, aconseguits,
        fallits);

END EST_ActualitzaAconseguitsAny;
/

/*****************************************************************************
** Procediment EST_ActualitzaDespesaUsr (quant, rsp)                        **
******************************************************************************
** Entrada:                                                                 **
**  quant (DECIMAL): Quantitat que el sistema ha gestionat (pagament o dev) **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Actualitza el mdul estadstic. Actualitza la mitja de aportaci real    **
** per usuari en els regals. Per a realitzar el clcul de la mitja, per     **
** defecte t en compte el total de participants en regals, doncs a priori  **
** s una informaci ms rellevant, al haver-hi usuaris que no participen.  **
** Per altra banda s'inclou codi per a que tingui en compte, si es desitja, **
** el total d'usuaris de la xarxa social, en comptes dels participants.     **
** Per motius d'eficincia s'emmagatzema el total aportat fins al moment,   **
** per evitar recalcular totes les despeses a cada nova transacci.         **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE EST_ActualitzaDespesaUsr (quant IN DECIMAL, rsp OUT VARCHAR2)
AS

  usuaris INT;

BEGIN

  rsp := 'OK';
  
  --Recopilaci del total d'usuaris. Es pot escollir entre el total de la xarxa
  --social o el total de participants. Donat que hi pot haver un gran nombre
  --d'usuaris no actius, per defecte es recupera el total de PARTICIPANTS.
  
    --Recopilem el nombre total d'usuaris del sistema
  --SELECT COUNT(*) INTO usuaris
  --FROM Usuari;
  
    --Recopilem el nombre total de participants del sistema (per defecte)
  SELECT COUNT(DISTINCT participant) INTO usuaris
  FROM Transaccio;
  
  --Inserim una nova linia amb l'any actual i la mitja actual, que al ser un
  --sol participant, s la seva despesa entre el total d'usuaris.
  INSERT INTO Regal_Any(any_reg, aux_despesa, mitja_despesa)
    VALUES(EXTRACT(YEAR FROM SYSDATE), quant, (quant/usuaris));

  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
    VALUES (id_log.NextVal, 'EST_ActualitzaDespesaUsr', rsp, quant);

EXCEPTION

  --Si l'any ja existeix, actualitzem les dades.
  WHEN DUP_VAL_ON_INDEX THEN
  --Primer actualitzem la despesa total realitzada al sistema en l'any actual
  UPDATE Regal_Any r
    SET r.aux_despesa = r.aux_despesa + quant
    WHERE r.any_reg = EXTRACT(YEAR FROM SYSDATE);
  --Finalment, actualitzem la mitja de despesa de l'any
  UPDATE Regal_Any r
    SET r.mitja_despesa = r.aux_despesa / usuaris
    WHERE r.any_reg = EXTRACT(YEAR FROM SYSDATE);
    
  --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
      VALUES (id_log.NextVal, 'EST_ActualitzaDespesaUsr', rsp, quant);

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
      VALUES (id_log.NextVal, 'EST_ActualitzaDespesaUsr', rsp, quant);

END EST_ActualitzaDespesaUsr;
/

/*****************************************************************************
** Procediment EST_ActualitzaUsuariRegals (rsp)                             **
******************************************************************************
** Entrada:                                                                 **
**  El procediment no necessita parmetres d'entrada.                       **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Actualitza el mdul estadstic. Enregistra i mant actualitzada la       **
** la estadstica de l'usuari que ha rebut ms regals, i quants n'ha rebut. **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE EST_ActualitzaUsuariRegals (rsp OUT VARCHAR2)
AS

  usr_regals Regal_Any_Mes.usuari_mes_regals%TYPE;
  num_regals Regal_Any_Mes.aux_max_regals%TYPE;
  max_actuals Regal_Any_Mes.aux_max_regals%TYPE;
  
  --Cursor que retorna, en la primera posici, l'usuari amb ms regals rebuts
  --en el moment de la crida del procediment d'actualitzaci.
  --Quan un regal es rep, el desig consta com 'ACONSEGUIT'. Fem servir la taula
  --DESIG, doncs cont informaci de la data de consecuci del regal.
  CURSOR resultats IS
    SELECT usuari, COUNT (*) AS Rebuts
    FROM Desig
    WHERE estat = 'ACONSEGUIT'
      AND EXTRACT(YEAR FROM data_estat) = EXTRACT(YEAR FROM SYSDATE)
      AND EXTRACT(MONTH FROM data_estat) = EXTRACT(MONTH FROM SYSDATE)
    GROUP BY Usuari
    ORDER BY Rebuts desc;

BEGIN

  rsp := 'OK';
  --Desem l'usuari amb ms regals i quants n'ha rebut.
  OPEN resultats;
  FETCH resultats INTO usr_regals, num_regals;
  CLOSE resultats;
  
  --Si noms ha rebut un regal, caldr inicialitzar la entrada del mes actual.
  --En cas que ja estigui inicialitzada, la excepci s'encarrega de seguir.
  IF num_regals = 1 THEN
    INSERT INTO Regal_Any_Mes(any_reg, mes, aux_max_regals, usuari_mes_regals)
      VALUES(EXTRACT(YEAR FROM SYSDATE), EXTRACT(MONTH FROM SYSDATE),
        num_regals, usr_regals);
  --Si hi ha ms d'un regal, hem de mirar si s ms gran que el mxim actual.
  ELSE
    SELECT r.aux_max_regals INTO max_actuals
    FROM Regal_Any_Mes r
    WHERE r.any_reg = EXTRACT(YEAR FROM SYSDATE)
      AND r.mes =  EXTRACT(MONTH FROM SYSDATE);
    --Si s ms gran, actualitzem les dades de la taula del mdul.
    IF num_regals > max_actuals THEN
      UPDATE Regal_Any_Mes r
      SET r.aux_max_regals = num_regals
      WHERE any_reg = EXTRACT(YEAR FROM SYSDATE)
        AND mes =  EXTRACT(MONTH FROM SYSDATE);
      UPDATE Regal_Any_Mes r
      SET r.usuari_mes_regals = usr_regals
      WHERE any_reg = EXTRACT(YEAR FROM SYSDATE)
        AND mes =  EXTRACT(MONTH FROM SYSDATE);
    END IF;
  END IF;
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1)
    VALUES (id_log.NextVal, 'EST_ActualitzaUsuariRegals', rsp);
  
EXCEPTION

  WHEN DUP_VAL_ON_INDEX THEN
    --Si s'ha rebut un regal i la taula estava inicialitzada, no cal fer res,
    --doncs el nou usuari tindr els mateixos regals que l'actual, i no ms.
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1)
      VALUES (id_log.NextVal, 'EST_ActualitzaUsuariRegals', rsp);

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1)
      VALUES (id_log.NextVal, 'EST_ActualitzaUsuariRegals', rsp);

END EST_ActualitzaUsuariRegals;
/

/*****************************************************************************
** Procediment EST_ActualitzaUsrMaxDespesa (rsp)                            **
******************************************************************************
** Entrada:                                                                 **
**  El procediment no necessita parmetres d'entrada.                       **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Actualitza el mdul estadstic. Enregistra i mant actualitzada la       **
** la estadstica de l'usuari amb ms despesa realitzada en regals.         **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE EST_ActualitzaUsrMaxDespesa (rsp OUT VARCHAR2)
AS

  usr_regals Regal_Any_Mes.usuari_mes_despesa%TYPE;
  usr_despesa Regal_Any_Mes.aux_max_despesa_usr%TYPE;
  max_despesa Regal_Any_Mes.aux_max_despesa_usr%TYPE;
  
  --Cursor que retorna, en la primera posici, l'usuari amb ms aportaci
  --en regals en el moment de la crida del procediment d'actualitzaci.
  CURSOR resultats IS
    SELECT participant, SUM(quantitat) As Despesa
    FROM Transaccio
    WHERE EXTRACT(YEAR FROM instant) = EXTRACT(YEAR FROM SYSDATE)
      AND EXTRACT(MONTH FROM instant) = EXTRACT(MONTH FROM SYSDATE)
    GROUP BY participant
    ORDER BY Despesa desc;

BEGIN

  rsp := 'OK';
  --Desem l'usuari amb ms regals i quants n'ha rebut.
  OPEN resultats;
  FETCH resultats INTO usr_regals, usr_despesa;
  CLOSE resultats;
  
  --Com no sabem si la entrada de la taula corresponent al mes i any est creada
  --intentem afegir les dades. Si falla, la excepci se n'encarregar.
  --Si no est creada vol dir que s la primera participaci del mes i, per
  --tant, tamb s l'usuari amb ms despesa fins el moment.
  INSERT INTO Regal_Any_Mes(any_reg, mes, usuari_mes_despesa, aux_max_despesa_usr)
    VALUES(EXTRACT(YEAR FROM SYSDATE), EXTRACT(MONTH FROM SYSDATE),
      usr_regals, usr_despesa);

  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1)
    VALUES (id_log.NextVal, 'EST_ActualitzaUsrMaxDespesa', rsp);
  
EXCEPTION

  WHEN DUP_VAL_ON_INDEX THEN
  --Si hi ha una despesa mxima, hem de verificar de nou les dades. Si el valor
  --s idntic, no cal actualitzar, doncs un altre usuari ha assolit la mateixa
  --quantitat. Si s diferent cal actualitzar, doncs poden haver devolucions
  --en el sistema i haver-se reduit la cuantitat mxima de despesa.
    SELECT r.aux_max_despesa_usr INTO max_despesa
    FROM Regal_Any_Mes r
    WHERE r.any_reg = EXTRACT(YEAR FROM SYSDATE)
      AND r.mes =  EXTRACT(MONTH FROM SYSDATE);
    IF usr_despesa != max_despesa THEN
      UPDATE Regal_Any_Mes r
      SET r.aux_max_despesa_usr = usr_despesa
      WHERE any_reg = EXTRACT(YEAR FROM SYSDATE)
        AND mes =  EXTRACT(MONTH FROM SYSDATE);
      UPDATE Regal_Any_Mes r
      SET r.usuari_mes_despesa = usr_regals
      WHERE any_reg = EXTRACT(YEAR FROM SYSDATE)
        AND mes =  EXTRACT(MONTH FROM SYSDATE);
    END IF;

    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1)
      VALUES (id_log.NextVal, 'EST_ActualitzaUsrMaxDespesa', rsp);

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1)
      VALUES (id_log.NextVal, 'EST_ActualitzaUsrMaxDespesa', rsp);

END EST_ActualitzaUsrMaxDespesa;
/

/*****************************************************************************
** Procediment EST_ActualitzaUsrMaxPart (rsp)                               **
******************************************************************************
** Entrada:                                                                 **
**  El procediment no necessita parmetres d'entrada.                       **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Actualitza el mdul estadstic. Enregistra i mant actualitzada la       **
** la estadstica de l'usuari amb ms participacions en regals. Cal tenir   **
** en compte que, al tractar-se de participacions, s'ignora el fet que un   **
** regal no s'hagi assolit: es compta el fet de participar en s.           **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE EST_ActualitzaUsrMaxPart (rsp OUT VARCHAR2)
AS

  usr_regals Regal_Any_Mes.usuari_mes_participacions%TYPE;
  usr_participacions Regal_Any_Mes.aux_max_participacions%TYPE;
  max_participacions Regal_Any_Mes.aux_max_participacions%TYPE;
 
  
  --Cursor que retorna, en la primera posici, l'usuari amb ms participacions
  --en regals en el moment de la crida del procediment d'actualitzaci.
  CURSOR resultats IS    
    --Comptem el nombre de regals...
    SELECT participant, COUNT (regal) AS Participacions
    FROM
      --...de la llista de regals en que ha participat cada usuari
      (SELECT participant, titol_hom, usuari_hom, regal
      FROM transaccio
      WHERE EXTRACT(YEAR FROM instant) = EXTRACT(YEAR FROM SYSDATE)
        AND EXTRACT(MONTH FROM instant) = EXTRACT(MONTH FROM SYSDATE)
      GROUP BY participant, titol_hom, usuari_hom, regal
      ORDER BY participant)
    GROUP BY participant
    ORDER BY participacions desc;

BEGIN

  rsp := 'OK';
  --Desem l'usuari amb ms participacions i quantes n'ha realitzat.
  OPEN resultats;
  FETCH resultats INTO usr_regals, usr_participacions;
  CLOSE resultats;
  
  --Com no sabem si la entrada de la taula corresponent al mes i any est creada
  --intentem afegir les dades. Si falla, la excepci se n'encarregar.
  --Si no est creada vol dir que s la primera participaci del mes i, per
  --tant, tamb s l'usuari amb ms participacions fins el moment.
  INSERT INTO Regal_Any_Mes(any_reg, mes, usuari_mes_participacions, aux_max_participacions)
    VALUES(EXTRACT(YEAR FROM SYSDATE), EXTRACT(MONTH FROM SYSDATE),
      usr_regals, usr_participacions);

  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1)
    VALUES (id_log.NextVal, 'EST_ActualitzaUsrMaxPart', rsp);
  
EXCEPTION

  WHEN DUP_VAL_ON_INDEX THEN
  --Si ja hi han participacions, hem de verificar si el nombre s major que
  --l'actual i actualitzar en cas necessari.
    SELECT r.aux_max_participacions INTO max_participacions
    FROM Regal_Any_Mes r
    WHERE r.any_reg = EXTRACT(YEAR FROM SYSDATE)
      AND r.mes =  EXTRACT(MONTH FROM SYSDATE);
    --Si s ms gran, actualitzem les dades de la taula del mdul.
    IF usr_participacions > max_participacions THEN
      UPDATE Regal_Any_Mes r
      SET r.aux_max_participacions = usr_participacions
      WHERE any_reg = EXTRACT(YEAR FROM SYSDATE)
        AND mes =  EXTRACT(MONTH FROM SYSDATE);
      UPDATE Regal_Any_Mes r
      SET r.usuari_mes_participacions = usr_regals
      WHERE any_reg = EXTRACT(YEAR FROM SYSDATE)
        AND mes =  EXTRACT(MONTH FROM SYSDATE);
    END IF;

    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1)
      VALUES (id_log.NextVal, 'EST_ActualitzaUsrMaxPart', rsp);

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1)
      VALUES (id_log.NextVal, 'EST_ActualitzaUsrMaxPart', rsp);

END EST_ActualitzaUsrMaxPart;
/

/*****************************************************************************
** Procediment EST_ActualitzaGrupDespesa (rsp)                              **
******************************************************************************
** Entrada:                                                                 **
**  El procediment no necessita parmetres d'entrada.                       **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Actualitza el mdul estadstic. Enregistra i mant actualitzada la       **
** la estadstica del grup amb ms despesa en regals.                       **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE EST_ActualitzaGrupDespesa (rsp OUT VARCHAR2)
AS

  grup_regals Regal_Any_Mes.nom_grup_mes_despesa%TYPE;
  hom_regals Regal_any_Mes.hom_grup_mes_despesa%TYPE;
  grup_despesa Regal_Any_Mes.aux_max_despesa_grup%TYPE;
  max_despesa Regal_Any_Mes.aux_max_despesa_grup%TYPE;

  --Cursor que retorna, en la primera posici, el grup amb ms aportaci
  --en regals en el moment de la crida del procediment d'actualitzaci.
  CURSOR resultats IS
    SELECT titol_hom, usuari_hom, SUM(quantitat) As Despesa
    FROM Transaccio
    WHERE EXTRACT(YEAR FROM instant) = EXTRACT(YEAR FROM SYSDATE)
      AND EXTRACT(MONTH FROM instant) = EXTRACT(MONTH FROM SYSDATE)
    GROUP BY titol_hom, usuari_hom
    ORDER BY Despesa desc;

BEGIN

  rsp := 'OK';
  --Desem el grup amb ms aportaci.
  OPEN resultats;
  FETCH resultats INTO grup_regals, hom_regals, grup_despesa;
  CLOSE resultats;
  
  --Com no sabem si la entrada de la taula corresponent al mes i any est creada
  --intentem afegir les dades. Si falla, la excepci se n'encarregar.
  --Si no est creada vol dir que s la primera transacci del mes i, per
  --tant, tamb s el grup amb ms despesa fins el moment.
  INSERT INTO Regal_Any_Mes(any_reg, mes, nom_grup_mes_despesa,
    hom_grup_mes_despesa, aux_max_despesa_grup)
    VALUES(EXTRACT(YEAR FROM SYSDATE), EXTRACT(MONTH FROM SYSDATE),
      grup_regals, hom_regals, grup_despesa);

  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1)
    VALUES (id_log.NextVal, 'EST_ActualitzaGrupDespesa', rsp);
  
EXCEPTION

  WHEN DUP_VAL_ON_INDEX THEN
  --Si hi ha una despesa mxima, hem de verificar de nou les dades. Si el valor
  --s idntic, no cal actualitzar, doncs un altre grup ha assolit la mateixa
  --quantitat. Si s diferent cal actualitzar, doncs poden haver devolucions
  --en el sistema i haver-se reduit la cuantitat mxima de despesa.
    SELECT r.aux_max_despesa_grup INTO max_despesa
    FROM Regal_Any_Mes r
    WHERE r.any_reg = EXTRACT(YEAR FROM SYSDATE)
      AND r.mes =  EXTRACT(MONTH FROM SYSDATE);
    --Si s ms gran, actualitzem les dades de la taula del mdul.
    IF grup_despesa != max_despesa THEN
      UPDATE Regal_Any_Mes r
      SET r.aux_max_despesa_grup = grup_despesa
      WHERE any_reg = EXTRACT(YEAR FROM SYSDATE)
        AND mes =  EXTRACT(MONTH FROM SYSDATE);
      UPDATE Regal_Any_Mes r
      SET r.nom_grup_mes_despesa = grup_regals
      WHERE any_reg = EXTRACT(YEAR FROM SYSDATE)
        AND mes =  EXTRACT(MONTH FROM SYSDATE);
      UPDATE Regal_Any_Mes r
      SET r.hom_grup_mes_despesa = hom_regals
      WHERE any_reg = EXTRACT(YEAR FROM SYSDATE)
        AND mes =  EXTRACT(MONTH FROM SYSDATE);
    END IF;

    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1)
      VALUES (id_log.NextVal, 'EST_ActualitzaGrupDespesa', rsp);

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1)
      VALUES (id_log.NextVal, 'EST_ActualitzaGrupDespesa', rsp);

END EST_ActualitzaGrupDespesa;
/

-----------------PROCEDIMENTS DE CONSULTA DEL MDUL ESTADSTIC-----------------

/*****************************************************************************
** Procediment EST_VegadesProducte (prod, any_r, rsp)                       **
******************************************************************************
** Entrada:                                                                 **
**  prod (VARCHAR2): Producte que es vol consultar                          **
**  d_any (SMALLINT): Any que es vol consultar                              **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Realitza una consulta al mdul estadstic per obtenir quantes vegades    **
** s'ha regalat un producte (independentment del proveidor) en un any       **
** denmanat.                                                                **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE EST_VegadesProducte (prod IN VARCHAR2, any_r IN VARCHAR2,
  rsp OUT VARCHAR2)
AS

  cops Regal_Vegades.vegades%TYPE;

BEGIN

  rsp := 'OK';
  
  --Recuperem la quantitat de vegades que s'ha adquirit un producte
  SELECT r.vegades INTO cops
  FROM Regal_Vegades r
  WHERE r.producte = prod AND r.any_reg = any_r;

  --I mostrem les dades
  DBMS_OUTPUT.PUT_LINE('Vegades que s''ha regalat ' || prod || ' durant l''any '
    || any_r || ': ' || cops || '.');

  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
    VALUES (id_log.NextVal, 'EST_VegadesProducte', rsp, prod, any_r);

EXCEPTION

  WHEN NO_DATA_FOUND THEN
    DBMS_OUTPUT.PUT_LINE('Vegades que s''ha regalat ' || prod || ' durant l''any '
     || any_r || ': ' || 0 || '.');

    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
      VALUES (id_log.NextVal, 'EST_VegadesProducte', rsp, prod, any_r);

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);

    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
      VALUES (id_log.NextVal, 'EST_VegadesProducte', rsp, prod, any_r);

END EST_VegadesProducte;
/

/*****************************************************************************
** Procediment EST_AdquiritsAny (any_r, rsp)                                **
******************************************************************************
** Entrada:                                                                 **
**  d_any (SMALLINT): Any que es vol consultar                              **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Realitza una consulta al mdul estadstic per obtenir quants regals      **
** s'han adquirit durant un any concret.                                    **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE EST_AdquiritsAny (any_r IN VARCHAR2, rsp OUT VARCHAR2)
AS

  totals Regal_Any.adquirits%TYPE;

BEGIN

  rsp := 'OK';
  
  --Carreguem el nombre de regals adquirits
  SELECT r.adquirits INTO totals
  FROM Regal_Any r
  WHERE r.any_reg = any_r;

  --I mostrem la informaci
  DBMS_OUTPUT.PUT_LINE('Nombre total de regals adquirits durant l''any '
    || any_r || ': ' || totals || '.');

  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
    VALUES (id_log.NextVal, 'EST_AdquiritsAny', rsp, any_r);

EXCEPTION

  WHEN NO_DATA_FOUND THEN
    DBMS_OUTPUT.PUT_LINE('Nombre total de regals adquirits durant l''any '
     || any_r || ': ' || 0 || '.');

    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
      VALUES (id_log.NextVal, 'EST_AdquiritsAny', rsp, any_r);

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);

    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
      VALUES (id_log.NextVal, 'EST_AdquiritsAny', rsp, any_r);

END EST_AdquiritsAny;
/


/*****************************************************************************
** Procediment EST_ImportAny (any_r, rsp)                                   **
******************************************************************************
** Entrada:                                                                 **
**  d_any (SMALLINT): Any que es vol consultar                              **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Realitza una consulta al mdul estadstic per obtenir l'import total     **
** acumulat en regals adquirits durant un any concret.                      **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE EST_ImportAny (any_r IN VARCHAR2, rsp OUT VARCHAR2)
AS

  acumulat Regal_Any.imp_acumulat%TYPE;

BEGIN

  rsp := 'OK';
  
  --Carreguem l'import acumulat en regals
  SELECT r.imp_acumulat INTO acumulat
  FROM Regal_Any r
  WHERE r.any_reg = any_r;

  --I mostrem les dades demanades
  DBMS_OUTPUT.PUT_LINE('Import acumulat total de regals adquirits durant l''any '
    || any_r || ': ' || acumulat || '.');

  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
    VALUES (id_log.NextVal, 'EST_ImportAny', rsp, any_r);

EXCEPTION

  WHEN NO_DATA_FOUND THEN
    DBMS_OUTPUT.PUT_LINE('Import acumulat total de regals adquirits durant l''any '
     || any_r || ': ' || 0 || '.');

    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
      VALUES (id_log.NextVal, 'EST_ImportAny', rsp, any_r);

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);

    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
      VALUES (id_log.NextVal, 'EST_ImportAny', rsp, any_r);

END EST_ImportAny;
/

/*****************************************************************************
** Procediment EST_PercentatgeAny (any_r, rsp)                              **
******************************************************************************
** Entrada:                                                                 **
**  d_any (SMALLINT): Any que es vol consultar                              **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Realitza una consulta al mdul estadstic per obtenir el percentatge     **
** de regals que han aconseguit els fons respecte els que no ho han fet.    **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE EST_PercentatgeAny (any_r IN VARCHAR2, rsp OUT VARCHAR2)
AS

  perc Regal_Any.perc_aconseguits%TYPE;

BEGIN

  rsp := 'OK';
  
  --Carreguem el percentatge de regals aconseguits
  SELECT r.perc_aconseguits INTO perc
  FROM Regal_Any r
  WHERE r.any_reg = any_r;

  --I el mostrem
  DBMS_OUTPUT.PUT_LINE('Percentatge de regals aconseguits respecte no aconseguits en l''any '
    || any_r || ': ' || perc || '%.');

  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
    VALUES (id_log.NextVal, 'EST_PercentatgeAny', rsp, any_r);

EXCEPTION

  WHEN NO_DATA_FOUND THEN
    DBMS_OUTPUT.PUT_LINE('Percentatge de regals aconseguits respecte no aconseguits en l''any '
     || any_r || ': N/D');
    DBMS_OUTPUT.PUT_LINE('  Encara no s''han gestionat regals.');

    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
      VALUES (id_log.NextVal, 'EST_PercentatgeAny', rsp, any_r);

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);

    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
      VALUES (id_log.NextVal, 'EST_PercentatgeAny', rsp, any_r);

END EST_PercentatgeAny;
/

/*****************************************************************************
** Procediment EST_DespesaUsr (any_r, rsp)                                  **
******************************************************************************
** Entrada:                                                                 **
**  d_any (SMALLINT): Any que es vol consultar                              **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Realitza una consulta al mdul estadstic obtenint la mitja de despesa   **
** en regals per usuari participant o usuari de la xarxa (segons s'hagi     **
** configurat EST_ActualitzaDespesaUsr.                                     **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE EST_DespesaUsr (any_r IN VARCHAR2, rsp OUT VARCHAR2)
AS

  despesa Regal_Any.mitja_despesa%TYPE;

BEGIN

  rsp := 'OK';
  
  --Carreguem la mitja de despesa per usuari
  SELECT r.mitja_despesa INTO despesa
  FROM Regal_Any r
  WHERE r.any_reg = any_r;

  --I mostrem les dades
  DBMS_OUTPUT.PUT_LINE('Despesa mitja per usuari en l''any '
    || any_r || ': ' || despesa || '.');

  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
    VALUES (id_log.NextVal, 'EST_DespesaUsr', rsp, any_r);

EXCEPTION

  WHEN NO_DATA_FOUND THEN
  DBMS_OUTPUT.PUT_LINE('Despesa mitja per usuari en l''any '
     || any_r || ': N/D');
    DBMS_OUTPUT.PUT_LINE('  Encara no s''han realitzat transaccions.');

    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
      VALUES (id_log.NextVal, 'EST_DespesaUsr', rsp, any_r);

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);

    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
      VALUES (id_log.NextVal, 'EST_DespesaUsr', rsp, any_r);

END EST_DespesaUsr;
/

/*****************************************************************************
** Procediment EST_UsuariRegals (any_r, mes_r, rsp)                         **
******************************************************************************
** Entrada:                                                                 **
**  any_r (SMALLINT): Any que es vol consultar                              **
**  mes_r (SMALLINT): Mes que es vol consultar                              **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Realitza una consulta al mdul estadstic obtenint l'identificador de    **
** l'usuari que ha rebut ms regals en l'any i ms demanats.                **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE EST_UsuariRegals (any_r SMALLINT, mes_r SMALLINT, rsp OUT VARCHAR2)
AS

  usu Regal_Any_Mes.usuari_mes_regals%TYPE;
  reg Regal_Any_Mes.aux_max_regals%TYPE;

BEGIN

  rsp := 'OK';
  
  --Carreguem l'usuari amb ms regals rebuts i la quantitat dels mateixos
  SELECT r.usuari_mes_regals, r.aux_max_regals INTO usu, reg
  FROM Regal_Any_Mes r
  WHERE r.any_reg = any_r AND r.mes = mes_r;

  --I mostrem la informaci
  DBMS_OUTPUT.PUT_LINE('Usuari amb ms regals durant el ' || mes_r || '/' || any_r
    || ': ' || usu || ' (' || reg || ' regals).');

  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
    VALUES (id_log.NextVal, 'EST_UsuariRegals', rsp, any_r, mes_r);

EXCEPTION

  WHEN NO_DATA_FOUND THEN
  DBMS_OUTPUT.PUT_LINE('Usuari amb ms regals durant el ' || mes_r || '/' || any_r
     || ': N/D');
    DBMS_OUTPUT.PUT_LINE('  Encara no s''han entregat regals.');

    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
      VALUES (id_log.NextVal, 'EST_UsuariRegals', rsp, any_r, mes_r);

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);

    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
      VALUES (id_log.NextVal, 'EST_UsuariRegals', rsp, any_r, mes_r);

END EST_UsuariRegals;
/

/*****************************************************************************
** Procediment EST_UsuariMaxDespesa (any_r, mes_r, rsp)                     **
******************************************************************************
** Entrada:                                                                 **
**  any_r (SMALLINT): Any que es vol consultar                              **
**  mes_r (SMALLINT): Mes que es vol consultar                              **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Realitza una consulta al mdul estadstic obtenint l'identificador de    **
** l'usuari que realitzat ms despesa en regals en l'any i mes indicats.    **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE EST_UsuariMaxDespesa (any_r SMALLINT, mes_r SMALLINT, rsp OUT VARCHAR2)
AS

  usu Regal_Any_Mes.usuari_mes_despesa%TYPE;
  desp Regal_Any_Mes.aux_max_despesa_usr%TYPE;

BEGIN

  rsp := 'OK';
  
  --Carreguem l'usuari amb ms despesa en regals i la quantitat de la mateixa
  SELECT r.usuari_mes_despesa, r.aux_max_despesa_usr INTO usu, desp
  FROM Regal_Any_Mes r
  WHERE r.any_reg = any_r AND r.mes = mes_r;

  --I mostrem les dades obtingudes
  DBMS_OUTPUT.PUT_LINE('Usuari amb ms despesa en regals durant el ' || mes_r ||
    '/' || any_r || ': ' || usu || ' (' || desp || ' ).');

  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
    VALUES (id_log.NextVal, 'EST_UsuariMaxDespesa', rsp, any_r, mes_r);

EXCEPTION

  WHEN NO_DATA_FOUND THEN
    DBMS_OUTPUT.PUT_LINE('Usuari amb ms despesa en regals durant el ' || mes_r ||
      '/' || any_r || ': N/D');
    DBMS_OUTPUT.PUT_LINE('  Encara no s''han realitzat transaccions.');

    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
      VALUES (id_log.NextVal, 'EST_UsuariMaxDespesa', rsp, any_r, mes_r);

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);

    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
      VALUES (id_log.NextVal, 'EST_UsuariMaxDespesa', rsp, any_r, mes_r);

END EST_UsuariMaxDespesa;
/

/*****************************************************************************
** Procediment EST_UsuariMaxPart (any_r, mes_r, rsp)                        **
******************************************************************************
** Entrada:                                                                 **
**  any_r (SMALLINT): Any que es vol consultar                              **
**  mes_r (SMALLINT): Mes que es vol consultar                              **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Realitza una consulta al mdul estadstic obtenint l'identificador de    **
** l'usuari que realitzat ms participacions en regals en l'any i mes       **
** indicats. Cal tenir en compte que es compta el fet de participar en un   **
** regal, independentment de quantes vegades s'hi participa o si no s'acon- **
** segueix a temps, s a dir, la intenci de l'usuari de participar.        **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE EST_UsuariMaxPart (any_r SMALLINT, mes_r SMALLINT, rsp OUT VARCHAR2)
AS

  usu Regal_Any_Mes.usuari_mes_participacions%TYPE;
  part Regal_Any_Mes.aux_max_participacions%TYPE;

BEGIN

  rsp := 'OK';
  
  --Carreguem l'usuari amb ms participacions i el nombre de les mateixes
  SELECT r.usuari_mes_participacions, r.aux_max_participacions INTO usu, part
  FROM Regal_Any_Mes r
  WHERE r.any_reg = any_r AND r.mes = mes_r;

  --I mostrem les dades demanades
  DBMS_OUTPUT.PUT_LINE('Usuari amb ms participaci en regals durant el ' || mes_r ||
    '/' || any_r || ': ' || usu || ' (' || part || ' participacions).');

  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
    VALUES (id_log.NextVal, 'EST_UsuariMaxPart', rsp, any_r, mes_r);

EXCEPTION

  WHEN NO_DATA_FOUND THEN
  DBMS_OUTPUT.PUT_LINE('Usuari amb ms participaci en regals durant el ' || mes_r ||
    '/' || any_r || ': N/D');
    DBMS_OUTPUT.PUT_LINE('  Encara no s''han realitzat participacions.');

    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
      VALUES (id_log.NextVal, 'EST_UsuariMaxPart', rsp, any_r, mes_r);

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);

    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
      VALUES (id_log.NextVal, 'EST_UsuariMaxPart', rsp, any_r, mes_r);

END EST_UsuariMaxPart;
/

/*****************************************************************************
** Procediment EST_GrupMaxDespesa (any_r, mes_r, rsp)                       **
******************************************************************************
** Entrada:                                                                 **
**  any_r (SMALLINT): Any que es vol consultar                              **
**  mes_r (SMALLINT): Mes que es vol consultar                              **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Realitza una consulta al mdul estadstic obtenint l'identificador del   **
** grup que ha realitzat ms despesa en regals en l'any i mes indicats.     **
** Cal recordar que un grup s'identifica pel seu nom i homenatjat.          **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE EST_GrupMaxDespesa (any_r SMALLINT, mes_r SMALLINT, rsp OUT VARCHAR2)
AS

  grup Regal_Any_Mes.nom_grup_mes_despesa%TYPE;
  hom Regal_Any_Mes.hom_grup_mes_despesa%TYPE;
  desp Regal_Any_Mes.aux_max_despesa_grup%TYPE;

BEGIN

  rsp := 'OK';
  
  --Carreguem el grup amb ms despesa i la quantitat de la mateixa
  SELECT r.nom_grup_mes_despesa, r.hom_grup_mes_despesa, r.aux_max_despesa_grup
    INTO grup, hom, desp
  FROM Regal_Any_Mes r
  WHERE r.any_reg = any_r AND r.mes = mes_r;

  --I mostrem les dades obtingudes
  DBMS_OUTPUT.PUT_LINE('Grup amb ms despesa en regals durant el ' || mes_r ||
    '/' || any_r || ':');
  DBMS_OUTPUT.PUT_LINE('  ' || grup || ' - ' || hom || ' (' || desp || ' ).');

  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
    VALUES (id_log.NextVal, 'EST_GrupMaxDespesa', rsp, any_r, mes_r);

EXCEPTION

  WHEN NO_DATA_FOUND THEN
  DBMS_OUTPUT.PUT_LINE('Grup amb ms despesa en regals durant el ' || mes_r ||
    '/' || any_r || ': N/D');
    DBMS_OUTPUT.PUT_LINE('  Encara no s''han realitzat transaccions.');

    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
      VALUES (id_log.NextVal, 'EST_GrupMaxDespesa', rsp, any_r, mes_r);

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);

    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
      VALUES (id_log.NextVal, 'EST_GrupMaxDespesa', rsp, any_r, mes_r);

END EST_GrupMaxDespesa;
/

--------------------- PROCEDIMENTS DE LA BASE DE DADES -----------------------

/*****************************************************************************
** Procediment Desig_Afegir(usr, prod, prov, rsp)                           **
******************************************************************************
** Entrada:                                                                 **
**  usr (VARCHAR2): Usuari al que es vol afegir el desig.                   **
**  prod (VARCHAR2): Producte que desitja l'usuari indicat,                 **
**  prov (VARCHAR2): Provedor que ofereix el producte.                     ** 
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Permet afegir un desig a la llista d'un usuari.                          **
** Genera un codi automtic de desig del tipus Num-Usuari, permetent dife-  **
** rents cpies d'un mateix desig, si l'usuari en desitja ms d'un.         **
*****************************************************************************/
CREATE OR REPLACE 
PROCEDURE Desig_Afegir(usr IN VARCHAR2, prod IN VARCHAR2, prov IN VARCHAR2,
  rsp OUT VARCHAR2)
AS
  num int;
  cod VARCHAR(50);
  
  DADES_INCORRECTES EXCEPTION;
  PRAGMA EXCEPTION_INIT(DADES_INCORRECTES, -2291);
  
BEGIN
  rsp := 'OK';
  --Es genera el codi del desig. Permet desitjar el mateix producte ms cops.
  SELECT COUNT (*) INTO num FROM Desig WHERE usuari = usr;
  cod := num + 1 || '-' || usr;
  --Inserim el desig
  INSERT INTO Desig (codi, producte, proveidor, usuari, data_pet, data_estat, estat)
    VALUES (cod, prod, prov, usr, SYSDATE, SYSDATE, 'PENDENT');
    
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3)
    VALUES (id_log.NextVal, 'Desig_Afegir', rsp, usr, prod, prov);

EXCEPTION

  WHEN DADES_INCORRECTES THEN
  rsp := 'ERROR: Usuari i/o producte incorrectes.';
  DBMS_OUTPUT.PUT_LINE(rsp);
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3)
    VALUES (id_log.NextVal, 'Desig_Afegir', rsp, usr, prod, prov);

  WHEN OTHERS THEN
  rsp := 'ERROR: ' || SQLERRM;
  DBMS_OUTPUT.PUT_LINE(rsp);
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3)
    VALUES (id_log.NextVal, 'Desig_Afegir', rsp, usr, prod, prov);
  
END Desig_Afegir;
/

/*****************************************************************************
** Procediment Desig_Consultar(usr, rsp)                                    **
******************************************************************************
** Entrada:                                                                 **
**  usr (VARCHAR2): Usuari al que es vol afegir el desig.                   **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Mostra la llista de desitjos actuals de l'usuari indicat.                **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE Desig_Consultar (usr IN VARCHAR2, rsp OUT VARCHAR2)
AS

  num DECIMAL (10,2);
  --Cursor que recupera els desitjos pendents de l'usuari demanat
  CURSOR resultats IS
    SELECT
      d.producte,
      d.proveidor,
      p.preu,
      p.impost,
      d.data_pet
    FROM Desig d LEFT JOIN Producte p
    ON d.producte = p.nom AND d.proveidor = p.proveidor
    WHERE d.usuari = usr AND (d.estat = 'PENDENT' OR d.estat = 'ASSIGNAT')
    ORDER BY d.producte, d.proveidor;

BEGIN

  rsp := 'OK';
  --Mostrem una capalera
  DBMS_OUTPUT.PUT_LINE('Llista de desitjos de lusuari ' || usr );
  DBMS_OUTPUT.PUT_LINE('');
  DBMS_OUTPUT.PUT_LINE(RPAD('Producte',30) || RPAD('Provedor',25) || 
    RPAD('Preu',13) || RPAD('Alta desig',20));
  --Mostrem cada desig a la llista de l'usuari
  FOR registre IN resultats LOOP
    num := registre.preu + (registre.preu * registre.impost / 100);
    DBMS_OUTPUT.PUT_LINE(
        RPAD(registre.producte,30) ||
        RPAD(registre.proveidor,18) ||
        to_char (num, '999G999D99') || RPAD(' ',11) ||
        registre.data_pet); 
  END LOOP;
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
    VALUES (id_log.NextVal, 'Desig_Consultar', rsp, usr);
  
EXCEPTION

  WHEN OTHERS THEN
  rsp := 'ERROR: ' || SQLERRM;
  DBMS_OUTPUT.PUT_LINE(rsp);
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
    VALUES (id_log.NextVal, 'Desig_Consultar', rsp, usr);
  
END Desig_Consultar;
/

/*****************************************************************************
** Procediment Desig_Historial(usr, rsp)                                    **
******************************************************************************
** Entrada:                                                                 **
**  usr (VARCHAR2): Usuari al que es vol afegir el desig.                   **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Mostra la llista complerta de desitjos de l'usuari indicat.              **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE Desig_Historial (usr IN VARCHAR2, rsp OUT VARCHAR2)
AS

  num DECIMAL (10,2);
  --Cursor que recupera tots els desitjos i estats de l'usuari demanat
  CURSOR resultats IS
    SELECT
      d.estat,
      d.producte,
      d.proveidor,
      p.preu,
      p.impost,
      d.data_pet
    FROM Desig d LEFT JOIN Producte p
    ON d.producte = p.nom AND d.proveidor = p.proveidor
    WHERE d.usuari = usr
    ORDER BY d.data_pet, d.estat, d.producte, d.proveidor;
    
BEGIN

  rsp := 'OK';
  --Mostrem una capalera
  DBMS_OUTPUT.PUT_LINE('Historial de desitjos de lusuari ' || usr );
  DBMS_OUTPUT.PUT_LINE('');
  DBMS_OUTPUT.PUT_LINE(RPAD('Estat',15)||RPAD('Producte',30)||
    RPAD('Provedor',25)||RPAD('Preu',13)||RPAD('Alta desig',20));
  --I mostrem cada desig de lusuari
  FOR registre IN resultats LOOP
    num := registre.preu + (registre.preu * registre.impost / 100);
    DBMS_OUTPUT.PUT_LINE(
        RPAD(registre.estat,15) ||
        RPAD(registre.producte,30) ||
        RPAD(registre.proveidor,18) ||
        to_char (num, '999G999D99') || RPAD(' ',11) ||
        registre.data_pet); 
  END LOOP;
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
    VALUES (id_log.NextVal, 'Desig_Historial', rsp, usr);
  
EXCEPTION

  WHEN OTHERS THEN
  rsp := 'ERROR: ' || SQLERRM;
  DBMS_OUTPUT.PUT_LINE(rsp);
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
    VALUES (id_log.NextVal, 'Desig_Historial', rsp, usr);

END Desig_Historial;
/

/*****************************************************************************
** Procediment Desig_Esborrar(usr, prod, prov, rsp)                         **
/*****************************************************************************
** Entrada:                                                                 **
**  usr (VARCHAR2): Usuari del que es vol esborrar el desig.                **
**  prod (VARCHAR2): Producte que es desitja.                               **
**  prov (VARCHAR2): Proveidor que ofereix el producte.                     **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Marca un desig d'un usuari com a cancellat. En cas que estigui repetit, **
** modifica el pendent, i si n'hi ha ms d'un, n'esborra el ms antic.      **
** Si el desig est assignat a un esdeveniment, el sistema el bloqueja per  **
** evitar problemes al grup que ja l'est adquirint.                        **
*****************************************************************************/
CREATE OR REPLACE 
PROCEDURE Desig_Esborrar(usr IN VARCHAR2, prod IN VARCHAR2, prov IN VARCHAR2,
  rsp OUT VARCHAR2)
IS

  estat_mod desig.estat%TYPE;
  codi_mod desig.codi%TYPE;

  --Cursor que recupera el desig ms antic i el seu estat
  CURSOR resultats IS
    SELECT d.codi, d.estat  FROM Desig d
    WHERE d.usuari = usr
      AND d.producte = prod
      AND d.proveidor = prov
    ORDER BY d.estat desc, d.data_pet asc;
  
  ESTAT_INCORRECTE EXCEPTION;
  DESIG_ASSIGNAT EXCEPTION;
  DESIG_INCORRECTE EXCEPTION;

BEGIN

  rsp := 'OK';
  --Cerquem el desig
  OPEN resultats;
  FETCH resultats INTO codi_mod, estat_mod;
  CLOSE resultats;
  --Si no hi ha codi, s'han introduit dades incorrectes
  IF codi_mod IS NULL THEN RAISE DESIG_INCORRECTE;
  END IF;
  --Si l'estat s pendent, el podem cancellar sense afectar a cap usuari
  IF estat_mod = 'PENDENT' THEN
    UPDATE Desig
    SET estat = 'CANCELLAT', data_estat = SYSDATE
    WHERE codi = codi_mod;
  --Si est+a assignat el bloqueja per evitar discrepancies en el grup
  ELSIF estat_mod = 'ASSIGNAT' THEN RAISE  DESIG_ASSIGNAT;
  ELSE
    -- Els desitjos aconseguits i els cancellats no es poden cancellar.
    RAISE ESTAT_INCORRECTE;
  END IF;

  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3)
    VALUES (id_log.NextVal, 'Desig_Esborrar', rsp, usr, prod, prov);

EXCEPTION

  WHEN ESTAT_INCORRECTE THEN
    rsp := 'ERROR: El desig no presenta un estat correcte (ja cancellat o aconseguit).';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3)
      VALUES (id_log.NextVal, 'Desig_Esborrar', rsp, usr, prod, prov);

  WHEN DESIG_ASSIGNAT THEN
    rsp := 'ERROR: El desig est assignat a un esdeveniment.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3)
      VALUES (id_log.NextVal, 'Desig_Esborrar', rsp, usr, prod, prov);

  WHEN DESIG_INCORRECTE THEN
    rsp := 'ERROR: El desig no existeix.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3)
      VALUES (id_log.NextVal, 'Desig_Esborrar', rsp, usr, prod, prov);
   
  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3)
      VALUES (id_log.NextVal, 'Desig_Esborrar', rsp, usr, prod, prov);

END Desig_Esborrar;
/

/*****************************************************************************
** Procediment Desig_Modificar(usr, prod_antic, prov_antic, prod_nou,       **
**    prov_nou, rsp)                                                        **
******************************************************************************
** Entrada:                                                                 **
**  usr (VARCHAR2): Usuari del que es vol modificar el desig.               **
**  prod_antic (VARCHAR2): Producte que es vol modificar.                   **
**  prov_antic (VARCHAR2): Proveidor del producte a modificar.              **
**  prod_nou (VARCHAR2): Producte que es vol desitjar.                      **
**  prov_nou (VARCHAR2): Proveidor del producte nou.                        **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Modifica qualsevol desig pendent de l'usuari. En cas que sigui repetit,  **
** modifica el ms antic.                                                   **
*****************************************************************************/
CREATE OR REPLACE 
PROCEDURE Desig_Modificar (usr IN VARCHAR2, prod_antic IN VARCHAR2,
  prov_antic IN VARCHAR2, prod_nou IN VARCHAR2, prov_nou IN VARCHAR2,
  rsp OUT VARCHAR2)
IS

  estat_mod desig.estat%TYPE;
  codi_mod desig.codi%TYPE;

  --Cursor que recupera el desig demanat i, si est repetit, el ms antic primer
  CURSOR resultats IS
    SELECT d.codi, d.estat  FROM Desig d
    WHERE d.usuari = usr
      AND d.estat = 'PENDENT'
      AND d.producte = prod_antic
      AND d.proveidor = prov_antic
    ORDER BY d.estat desc, d.data_pet asc;

  DESIG_INCORRECTE EXCEPTION;

BEGIN

  rsp := 'OK';
  --Cerquem el desig
  OPEN resultats;
  FETCH resultats INTO codi_mod, estat_mod;
  CLOSE resultats;
  --Si no hi ha codi, el desig demanat no existeix
  IF codi_mod IS NULL THEN RAISE DESIG_INCORRECTE;
  END IF;
  --Modifiquem el desig demanat i actualitzem la data del canvi d'estat
  UPDATE Desig
  SET data_estat = SYSDATE, producte = prod_nou, proveidor = prov_nou
  WHERE codi = codi_mod;
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4, ent5)
    VALUES (id_log.NextVal, 'Desig_Modificar', rsp, usr, prod_antic, prov_antic,
      prod_nou, prov_nou);
  
EXCEPTION

  WHEN DESIG_INCORRECTE THEN
    rsp := 'ERROR: El desig no existeix o no presenta un estat correcte.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4, ent5)
      VALUES (id_log.NextVal, 'Desig_Modificar', rsp, usr, prod_antic, prov_antic,
        prod_nou, prov_nou);
   
  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4, ent5)
      VALUES (id_log.NextVal, 'Desig_Modificar', rsp, usr, prod_antic, prov_antic,
        prod_nou, prov_nou);
  
END Desig_Modificar;
/

/*****************************************************************************
** Procediment Categoria_Llistar(cat, rsp)                                  **
******************************************************************************
** Entrada:                                                                 **
**  cat (VARCHAR2): Categoria a llistar.                                    **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Mostra els regals disponibles de la categoria indicada.                  **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE Categoria_Llistar (cat IN VARCHAR2, rsp OUT VARCHAR2)
AS

  num DECIMAL (10,2);
  --Cursor que recupera les dades dels productes en una categoria
  CURSOR resultats IS
    SELECT
      p.nom,
      p.proveidor,
      p.preu,
      p.impost
    FROM Producte p
    WHERE p.categoria = cat
    ORDER BY p.nom, p.proveidor;
    
BEGIN

  rsp := 'OK';
  --Mostrem una capalera
  DBMS_OUTPUT.PUT_LINE('Llista de regals disponibles en la categoria ' || cat );
  DBMS_OUTPUT.PUT_LINE('');
  DBMS_OUTPUT.PUT_LINE(RPAD('Producte',30)||RPAD('Provedor',25)||RPAD('Preu',13)||RPAD('Impostos',20));
  --Mostrem cada registre
  FOR registre IN resultats LOOP
    DBMS_OUTPUT.PUT_LINE(
        RPAD(registre.nom,30) ||
        RPAD(registre.proveidor,18) ||
        to_char (registre.preu, '999G999D99') || RPAD(' ',11) ||
        registre.impost || '%'); 
  END LOOP;
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
    VALUES (id_log.NextVal, 'Categoria_Llistar', rsp, cat);

EXCEPTION
   
  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
      VALUES (id_log.NextVal, 'Categoria_Llistar', rsp, cat);
    
END Categoria_Llistar;
/

/*****************************************************************************
** Procediment Grup_AltaParticipant(grup, hom, part, dedic, rsp)            **
******************************************************************************
** Entrada:                                                                 **
**  grup (VARCHAR2): Nom del grup de regals.                                **
**  hom (VARCHAR2): Homenatjat del grup.                                    **
**  part (VARCHAR2): Usuari que afegim al grup.                             **
**  dedic (VARCHAR2): Dedicatoria personal del participant.                 **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Dona d'alta un participant en un grup de participants en regals.         **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE Grup_AltaParticipant (grup IN VARCHAR2, hom IN VARCHAR2,
  part IN VARCHAR2, dedic IN VARCHAR2, rsp OUT VARCHAR2)
AS

  --Si les dades no son correctes ho detectar com a que la clau fornia no s
  --correcta. Ho capturem per tractar-ho d'una manera ms descriptiva
  INSERT_WRONG_EVENT EXCEPTION;
  PRAGMA EXCEPTION_INIT(INSERT_WRONG_EVENT, -2291);

BEGIN
  rsp := 'OK';
  --Insertem les dades demanades. Si son incorrectes, la excepci assignada
  --se n'encarregar automticament
  INSERT INTO Integrant (titol_grup, hom_grup, usu_grup, dedicatoria)
  VALUES (grup, hom, part, dedic);
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent_text)
    VALUES (id_log.NextVal, 'Grup_AltaParticipant', rsp, grup, hom, part, dedic);
  
EXCEPTION

  WHEN DUP_VAL_ON_INDEX THEN
    rsp := 'ERROR: Aquest usuari ja forma part del grup.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent_text)
      VALUES (id_log.NextVal, 'Grup_AltaParticipant', rsp, grup, hom, part, dedic);
  
  WHEN INSERT_WRONG_EVENT THEN
    rsp := 'ERROR: El grup i/o usuari especificats no existeixen.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent_text)
      VALUES (id_log.NextVal, 'Grup_AltaParticipant', rsp, grup, hom, part, dedic);
    
  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent_text)
      VALUES (id_log.NextVal, 'Grup_AltaParticipant', rsp, grup, hom, part, dedic);

END Grup_AltaParticipant;
/
/*****************************************************************************
** Procediment Grup_BaixaParticipant(grup, hom, part, rsp)                  **
******************************************************************************
** Entrada:                                                                 **
**  grup (VARCHAR2): Nom del grup de regals.                                **
**  hom (VARCHAR2): Homenatjat del grup.                                    **
**  part (VARCHAR2): Usuari que afegim al grup.                             **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Dona de baixa un participant en un grup de participants en regals. El    **
** sistema noms permet la baixa si no ha aportat import, per evitar        **
** molsties a la resta del grup.                                           **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE Grup_BaixaParticipant (tit IN VARCHAR2, hom IN VARCHAR2,
  part IN VARCHAR2, rsp OUT VARCHAR2)
AS

  vegades INT;
  existeix INT;
  
  USUARI_AMB_PARTICIPACIONS EXCEPTION;
  DADES_INCORRECTES EXCEPTION;

BEGIN

  rsp := 'OK';

  --Mirem si l'usuari est al grup.
  SELECT COUNT (*) INTO existeix
  FROM Integrant i
  WHERE i.titol_grup = tit AND i.hom_grup = hom AND i.usu_grup = part;
  --Si no hi apareix cap cop, vol dir que no pertany al grup
  IF existeix = 0 THEN RAISE DADES_INCORRECTES;
  END IF;
  
  --Comptem si ha participat en regals
  SELECT COUNT (*) INTO vegades
  FROM Participacio p
  WHERE p.titol_hom = tit AND p.usuari_hom = hom AND p.participant = part;
  --Si ha participat, no en permet la baixa, podria ocasionar problemes amb
  --regals aconseguits i, per tant, molsties a altres participants del grup
  IF vegades > 0 THEN RAISE USUARI_AMB_PARTICIPACIONS;
  END IF;
  --Si tot s correcte, l'esborrem del grup
  DELETE FROM Integrant i
  WHERE i.titol_grup = tit AND i.hom_grup = hom AND i.usu_grup = part;
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3)
    VALUES (id_log.NextVal, 'Grup_BaixaParticipant', rsp, tit, hom, part);

EXCEPTION

  WHEN USUARI_AMB_PARTICIPACIONS THEN
    rsp := 'ERROR: Aquest usuari t aportacions en regals.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3)
      VALUES (id_log.NextVal, 'Grup_BaixaParticipant', rsp, tit, hom, part);
    
  WHEN DADES_INCORRECTES THEN
    rsp := 'ERROR: Usuari i/o grup incorrectes.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3)
      VALUES (id_log.NextVal, 'Grup_BaixaParticipant', rsp, tit, hom, part);

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3)
      VALUES (id_log.NextVal, 'Grup_BaixaParticipant', rsp, tit, hom, part);

END Grup_BaixaParticipant;
/

/*****************************************************************************
** Procediment Grup_Alta(tit, hom, dedic, d_fons, d_regals, prv, rec,       **
**  adr, pob, cpost, rsp)                                                   **
******************************************************************************
** Entrada:                                                                 **
**  tit (VARCHAR2): Nom del grup de regals.                                 **
**  hom (VARCHAR2): Homenatjat del grup.                                    **
**  dedic (VARCHAR2): Dedicatoria general del grup.                         **
**  d_fons (VARCHAR2): Data limit per a recopilar fons.                     **
**  d_regals (VARCHAR2): Data limit de recepcio dels regals.                **
**  prv (VARCHAR2): Indicador de privacitat del grup.                       **
**  rec (VARCHAR2): Usuari receptor dels regals.                            **
**  adr (VARCHAR2): Adrea d'entrega.                                       **
**  pob (VARCHAR2): Poblaci d'entrega.                                     **
**  cpost (VARCHAR2): Codi postal.                                          **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Dona d'alta un grup de participants (o esdeveniment d'homenatge).        **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE Grup_Alta (tit IN VARCHAR2, hom IN VARCHAR2, dedic IN VARCHAR2,
  d_fons IN DATE, d_regals IN DATE, prv IN VARCHAR2, rec IN VARCHAR2,
  adr IN VARCHAR2, pob IN VARCHAR2, cpost IN VARCHAR2, rsp OUT VARCHAR2)
AS

  codId VARCHAR(2);
  
  DADES_INCORRECTES EXCEPTION;
  TEMPS_INSUFICIENT EXCEPTION;
  PRAGMA EXCEPTION_INIT(DADES_INCORRECTES, -2291);
  PRAGMA EXCEPTION_INIT(TEMPS_INSUFICIENT, -2290);
  
BEGIN

  rsp := 'OK';
  --Per eficincia, es separen els primers dos nombres del codi postal
  codId := SUBSTR(cpost,1,2);
  --Donem d'alta el grup
  INSERT INTO Grup (titol, homenatjat, dedicatoria, data_fons, data_regals,
  privacitat, receptor, adreca, poblacio, cp, cp_id)
  VALUES (tit, hom, dedic, d_fons, d_regals, prv, rec, adr, pob, cpost, codId);
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4, ent5,
    ent6, ent7, ent8, ent9, ent_text)
    VALUES (id_log.NextVal, 'Grup_Alta', rsp, tit, hom, d_fons, d_regals, prv,
      rec, adr, pob, cpost, dedic);
  
EXCEPTION

  --Si el grup ja existeix, la excepci se n'encarrega
  WHEN DUP_VAL_ON_INDEX THEN
    rsp := 'ERROR: Aquest grup ja existeix.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4, ent5,
      ent6, ent7, ent8, ent9, ent_text)
      VALUES (id_log.NextVal, 'Grup_Alta', rsp, tit, hom, d_fons, d_regals, prv,
        rec, adr, pob, cpost, dedic);
        
  WHEN DADES_INCORRECTES THEN
    rsp := 'ERROR: L''usuari homenatjat indicat no s correcte.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4, ent5,
      ent6, ent7, ent8, ent9, ent_text)
      VALUES (id_log.NextVal, 'Grup_Alta', rsp, tit, hom, d_fons, d_regals, prv,
        rec, adr, pob, cpost, dedic);
        
    WHEN TEMPS_INSUFICIENT THEN
    rsp := 'ERROR: Temps insuficient per a l''entrega de regals.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4, ent5,
      ent6, ent7, ent8, ent9, ent_text)
      VALUES (id_log.NextVal, 'Grup_Alta', rsp, tit, hom, d_fons, d_regals, prv,
        rec, adr, pob, cpost, dedic);
    
  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4, ent5,
      ent6, ent7, ent8, ent9, ent_text)
      VALUES (id_log.NextVal, 'Grup_Alta', rsp, tit, hom, d_fons, d_regals, prv,
        rec, adr, pob, cpost, dedic);
    
END Grup_Alta;
/

/*****************************************************************************
** Procediment Grup_AfegirRegal (prod, prov, tit_hom, usu_hom, rsp)         **
******************************************************************************
** Entrada:                                                                 **
**  prod (VARCHAR2): Producte a afegir.                                     **
**  prov (VARCHAR2): Proveidor del producte.                                **
**  tit (VARCHAR2): Nom del grup de regals.                                 **
**  hom (VARCHAR2): Homenatjat del grup.                                    **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Afegeix un desig a un grup com a regal, i el marca com a assignat a la   **
** llista de desitjos de l'homenatjat. En cas que el desig estigui repetit  **
** a la llista de l'usuari, en selecciona el ms antic.                     **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE Grup_AfegirRegal (prod IN VARCHAR2, prov IN VARCHAR2, 
  tit_hom IN VARCHAR2, usu_hom IN VARCHAR2, rsp OUT VARCHAR2)
AS

  cod_desig Desig.codi%TYPE;
  estat_desig Desig.estat%TYPE;
  preu_iva Producte.preu%TYPE;
  impost Producte.impost%TYPE;
  --Cursor que recupera les dades d'un desig per assignar-hi un regal
  CURSOR resultats IS
    SELECT d.codi, d.estat, p.preu, p.impost
    FROM Desig d JOIN Producte p
    ON d.producte = p.nom
      AND d.proveidor = p.proveidor
    WHERE d.usuari = usu_hom
      AND d.estat = 'PENDENT'
      AND d.producte = prod
      AND d.proveidor = prov
    ORDER BY d.data_pet asc;
    
  ESTAT_INCORRECTE EXCEPTION;
  DESIG_ASSIGNAT EXCEPTION;
  DESIG_INCORRECTE EXCEPTION;
    
BEGIN

  rsp := 'OK';
  --Primer troba el codi del desig. Si n'hi ha ms d'un, escull el ms antic.
  OPEN resultats;
  FETCH resultats INTO cod_desig, estat_desig, preu_iva, impost;
  CLOSE resultats;
  IF cod_desig IS NULL THEN RAISE DESIG_INCORRECTE;
  END IF;
  IF estat_desig = 'PENDENT' THEN
    --Calculem el preu amb impostos i el guardem. Optimitzar les participacions.
    preu_iva := preu_iva + (preu_iva * impost / 100);
    --Afegim el desig a la llista de regals.
    INSERT INTO Regal (desig, titol_hom, usuari_hom, preu, estat)
    VALUES (cod_desig, tit_hom, usu_hom, preu_iva, 'AFEGIT_A_UN_GRUP');
    --Marquem el desig com assignat.
    UPDATE Desig d
      SET d.estat = 'ASSIGNAT'
      WHERE d.codi = cod_desig;
  --Si el desig ja est assignat, no es pot assignar de nou
  ELSIF estat_desig = 'ASSIGNAT' THEN RAISE DESIG_ASSIGNAT;
  ELSE RAISE ESTAT_INCORRECTE;
  END IF;

  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4)
    VALUES (id_log.NextVal, 'Grup_AfegirRegal', rsp, prod, prov, tit_hom, usu_hom);

EXCEPTION

  WHEN DUP_VAL_ON_INDEX THEN
    rsp := 'ERROR: Regal ja afegit al grup.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4)
      VALUES (id_log.NextVal, 'Grup_AfegirRegal', rsp, prod, prov, tit_hom, usu_hom);
    
  WHEN ESTAT_INCORRECTE THEN
    rsp := 'ERROR: El desig no presenta un estat correcte.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4)
      VALUES (id_log.NextVal, 'Grup_AfegirRegal', rsp, prod, prov, tit_hom, usu_hom);

  WHEN DESIG_ASSIGNAT THEN
    rsp := 'ERROR: El desig est assignat a un altre esdeveniment.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4)
      VALUES (id_log.NextVal, 'Grup_AfegirRegal', rsp, prod, prov, tit_hom, usu_hom);

  WHEN DESIG_INCORRECTE THEN
    rsp := 'ERROR: Desig incorrecte o inexistent.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4)
      VALUES (id_log.NextVal, 'Grup_AfegirRegal', rsp, prod, prov, tit_hom, usu_hom);
    
  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4)
      VALUES (id_log.NextVal, 'Grup_AfegirRegal', rsp, prod, prov, tit_hom, usu_hom);
    
END Grup_AfegirRegal;
/

/*****************************************************************************
** Procediment Grup_Participacions(tit, hom, rsp)                           **
******************************************************************************
** Entrada:                                                                 **
**  tit (VARCHAR2): Nom del grup de regals.                                 **
**  hom (VARCHAR2): Homenatjat del grup de regals.                          **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Mostra les participacions dels integrants d'un grup especificat. Si la   **
** banderola de privacitat del grup est activa, noms indica si l'usuari   **
** ha participat o no.                                                      **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE Grup_Participacions (tit IN VARCHAR2, hom IN VARCHAR2, 
  rsp OUT VARCHAR2)
AS

  priv Grup.privacitat%TYPE;

  --Cursor que recupera les participacions d'un usuari en un producte
  CURSOR resultats IS
    SELECT
      d.producte,
      d.proveidor,
      p.participant,
      p.aportacio
    FROM Participacio p LEFT JOIN Desig d
    ON p.desig = d.codi
    WHERE p.titol_hom = tit AND p.usuari_hom = hom
    ORDER BY d.producte, d.proveidor;

  --Cursor que recupera la privacitat d'un grup
  CURSOR es_privat IS
    SELECT
      g.privacitat
    FROM
      Grup g
    WHERE titol = tit AND homenatjat = hom;

  GRUP_INEXISTENT EXCEPTION;

BEGIN

  rsp := 'OK';
  
  --Desem la privacitat
  OPEN es_privat;
  FETCH es_privat INTO priv;
  CLOSE es_privat;
  --Aprofitem la privacitat per verificar dades. Si no existeix, el grup tampoc
  IF priv IS NULL THEN RAISE GRUP_INEXISTENT;
  END IF;
  
  --Mostrem una capalera
  DBMS_OUTPUT.PUT_LINE('Participacions del grup ' || tit );
  DBMS_OUTPUT.PUT_LINE('');
  DBMS_OUTPUT.PUT_LINE(RPAD('Producte',30) || RPAD('Provedor',25) || 
    RPAD('Participant',39) || RPAD('Aportacio',13));
  --Si no hi ha privacitat, mostrem totes les dades
  IF priv = 'F' THEN
    FOR registre IN resultats LOOP
      DBMS_OUTPUT.PUT_LINE(
        RPAD(registre.producte,30) ||
        RPAD(registre.proveidor,25) ||
        RPAD(registre.participant,35) ||
        to_char (registre.aportacio, '999G999D99') ||
        RPAD(' ',11));
    END LOOP;
  --Si hi ha privacitat, ocultem l'import de les participacions
  ELSE
    FOR registre IN resultats LOOP
      DBMS_OUTPUT.PUT_LINE(
        RPAD(registre.producte,30) ||
        RPAD(registre.proveidor,25) ||
        RPAD(registre.participant,40) ||
        RPAD('-Privat-',10)); 
    END LOOP;
  END IF;
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
    VALUES (id_log.NextVal, 'Grup_Participacions', rsp, tit, hom);
  
EXCEPTION

  WHEN GRUP_INEXISTENT THEN
    rsp := 'ERROR: El grup indicat no existeix.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
      VALUES (id_log.NextVal, 'Grup_Participacions', rsp, tit, hom);

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
      VALUES (id_log.NextVal, 'Grup_Participacions', rsp, tit, hom);

END Grup_Participacions;
/

/*****************************************************************************
** Procediment Grup_Consultar(tit, hom, rsp)                                **
******************************************************************************
** Entrada:                                                                 **
**  tit (VARCHAR2): Nom del grup de regals.                                 **
**  hom (VARCHAR2): Homenatjat del grup de regals.                          **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Mostra els integrants del grup demanat i l'estat dels regals             **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE Grup_Consultar (tit IN VARCHAR2, hom IN VARCHAR2, rsp OUT VARCHAR2)
AS

  priv Grup.privacitat%TYPE;
  inactiu Grup.finalitzat%TYPE;
  preu Regal.preu%TYPE;

  --Cursor que recupera la llista d'integrants d'un grup de regals
  CURSOR integrants IS
    SELECT
      i.usu_grup,
      i.dedicatoria
    FROM Integrant i
    WHERE i.titol_grup = tit AND i.hom_grup = hom
    ORDER BY usu_grup, dedicatoria asc;
    
  --Cursor que recupera la llista de regals d'un grup de regals
  CURSOR llista_regals IS
    SELECT
      d.producte,
      d.proveidor,
      r.preu,
      r.estat
    FROM Regal r LEFT JOIN Desig d
    ON r.desig = d.codi
    WHERE r.titol_hom = tit AND r.usuari_hom = hom
    ORDER BY d.producte asc, d.proveidor asc, r.preu asc;
      
  --Cursor que recupera la privacitat d'un grup
  CURSOR es_privat IS
    SELECT
      g.privacitat
    FROM
      Grup g
    WHERE titol = tit AND homenatjat = hom;
    
  --Cursor que recupera l'estat d'alta o baixa d'un grup de regals
  CURSOR es_actiu IS
    SELECT
      g.finalitzat
    FROM
      Grup g
    WHERE titol = tit AND homenatjat = hom;

  GRUP_INEXISTENT EXCEPTION;

BEGIN

  rsp := 'OK';
  
  --Desem la privacitat
  OPEN es_privat;
  FETCH es_privat INTO priv;
  CLOSE es_privat;
  
  --Si no existeix, les dades del grup sn incorrectes
  IF priv IS NULL THEN RAISE GRUP_INEXISTENT;
  ELSE
    --Mirem si el grup est d'alta
    OPEN es_actiu;
    FETCH es_actiu INTO inactiu;
    CLOSE es_actiu;
    --Mostrem una capalera
    DBMS_OUTPUT.PUT_LINE('');
    DBMS_OUTPUT.PUT_LINE('Grup de regals: ' || tit);
    DBMS_OUTPUT.PUT_LINE('En homenatge a: ' || hom);
    --I mostrem les dades del grup, adaptant-les segons sigui necessari
    IF priv = 'T' THEN
      DBMS_OUTPUT.PUT_LINE('Les aportacions dels membres del grup sn PRIVADES.');
    ELSE
      DBMS_OUTPUT.PUT_LINE('Les aportacions dels membres del grup sn PBLIQUES.');
    END IF;
    IF inactiu = 'T' THEN
      DBMS_OUTPUT.PUT_LINE('Aquest esdeveniment ha finalitzat.');
      DBMS_OUTPUT.PUT_LINE('');
    ELSE
      DBMS_OUTPUT.PUT_LINE('');
    END IF;
  END IF;
  
  DBMS_OUTPUT.PUT_LINE('Participants del grup:');
  DBMS_OUTPUT.PUT_LINE(RPAD(' Participant',40)|| ' Dedicatria personal');
  FOR registre IN integrants LOOP
      DBMS_OUTPUT.PUT_LINE('  ' || RPAD(registre.usu_grup,40) ||
        RPAD(registre.dedicatoria, 150));
  END LOOP;
  
  DBMS_OUTPUT.PUT_LINE('');
  DBMS_OUTPUT.PUT_LINE('Regals en la llista del grup:');
  DBMS_OUTPUT.PUT_LINE(RPAD('        Preu',21) || RPAD(' Regal',40) ||
    RPAD(' Proveidor',50));
  FOR registre IN llista_regals  LOOP
    DBMS_OUTPUT.PUT_LINE(to_char (registre.preu, '999G999D99') || 
      RPAD(' ',10) || RPAD(registre.producte,40) ||
      RPAD(registre.proveidor, 50) || RPAD(registre.estat, 20));
  END LOOP;
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
    VALUES (id_log.NextVal, 'Grup_Consultar', rsp, tit, hom);
  
EXCEPTION

  WHEN GRUP_INEXISTENT THEN
    rsp := 'ERROR: El grup indicat no existeix.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
      VALUES (id_log.NextVal, 'Grup_Consultar', rsp, tit, hom);

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
      VALUES (id_log.NextVal, 'Grup_Consultar', rsp, tit, hom);
  
END Grup_Consultar;
/

/*****************************************************************************
** Procediment Transaccio_Afegir (tipus, participant, titol_hom, usuari_hom,**
**    regal, dest_regal, mode_p, dades_p, quantitat, rsp)                   **
******************************************************************************
** Entrada:                                                                 **
**  tipus (VARCHAR2): Indica si la transacci s un pagament o devoluci.   **
**  participant (VARCHAR2): Usuari que participa en el regal.               **
**  titol_hom (VARCHAR2): Grup de regals on es participa.                   **
**  usuari_hom (VARCHAR2): Usuari homenatjat del grup.                      **
**  regal (VARCHAR2): Codi del regal.                                       **
**  dest_regal (VARCHAR2): Destinatari del regal.                           **
**  mode_p (VARCHAR2): Forma de pagament de la transacci.                  **
**  dades_p (VARCHAR2): Info del mode de pagament (compte PayPal, etc).     **
**  quantitat (DECIMAL): Quantitat que s'ha pagat/retornat.                 **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Afegeix una transacci a la llista de transaccions del sistema.          **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE Transaccio_Afegir (tipus IN VARCHAR2, participant IN VARCHAR2, 
  titol_hom IN VARCHAR2, usuari_hom IN VARCHAR2, regal IN VARCHAR2,
  dest_regal IN VARCHAR2, mode_p IN VARCHAR2, dades_p IN VARCHAR2,
  quantitat IN DECIMAL, rsp OUT VARCHAR2)
AS

  quant DECIMAL (10,2);
  
BEGIN

  rsp := 'OK';
  --Si s un pagament, la quantitat s positiva
  IF tipus = 'Pagament' THEN quant := quantitat;
    --Si s una devoluci, la quantitat s negativa
    ELSE quant := -quantitat;
  END IF;
  --Desem la transacci. if_tr genera un nmero de transacci consecutiu
  INSERT INTO Transaccio (id_tr, instant, tipus, participant, titol_hom,
    usuari_hom, regal, dest_regal, mode_p, dades_p, quantitat)
  VALUES (id_trans.NextVal, SYSDATE, tipus, participant, titol_hom, usuari_hom,
    regal, dest_regal, mode_p, dades_p, quant);
    
  --Actualitzar mdul estadstic
    --Actualitzar la mitjana de despesa per usuari
  EST_ActualitzaDespesaUsr (quant, rsp);
    --Actualitzar l'usuari amb ms despesa en regals
  EST_ActualitzaUsrMaxDespesa (rsp);
    --Actualitzar el grup amb ms despesa en regals
  EST_ActualitzaGrupDespesa (rsp);
    
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4,
    ent5, ent6, ent7, ent8, ent9)
    VALUES (id_log.NextVal, 'Transaccio_Afegir', rsp, tipus, participant,
      titol_hom, usuari_hom, regal, dest_regal, mode_p, dades_p, quantitat);
  
EXCEPTION

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4,
      ent5, ent6, ent7, ent8, ent9)
      VALUES (id_log.NextVal, 'Transaccio_Afegir', rsp, tipus, participant,
        titol_hom, usuari_hom, regal, dest_regal, mode_p, dades_p, quantitat);
    
END Transaccio_Afegir;
/

/*****************************************************************************
** Procediment Transaccio_Consultar(rsp)                                    **
******************************************************************************
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Mostra la llista de transaccions del sistema.                            **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE Transaccio_Mostrar (rsp OUT VARCHAR2)
AS

  --Cursor que recupera les transaccions del sistema
  CURSOR tr IS
    SELECT *
    FROM Transaccio t LEFT JOIN Desig d
    ON d.codi = t.regal
    ORDER BY id_tr;
    
BEGIN

  rsp := 'OK';
  --Mostrem una capalera
  DBMS_OUTPUT.PUT_LINE('Llista de transaccions del sistema ');
  DBMS_OUTPUT.PUT_LINE('');
  --I mostrem la llista de transaccions
  FOR reg IN tr LOOP
    DBMS_OUTPUT.PUT_LINE(
      RPAD(reg.id_tr,6) ||
      RPAD(reg.instant,10) ||
      RPAD(reg.tipus,10) ||
      RPAD(reg.participant,40) ||
      RPAD(reg.titol_hom,50) ||
      RPAD(reg.usuari_hom,40) ||
      RPAD(reg.producte,40) ||
      RPAD(reg.proveidor,40) ||
      RPAD(reg.dest_regal,40) ||
      RPAD(reg.mode_p,30) ||
      RPAD(reg.dades_p,40) ||
      to_char(reg.quantitat, '999G999D99') || RPAD(' ',10)
      );
  END LOOP;
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1)
    VALUES (id_log.NextVal, 'Transaccio_Mostrar', rsp);

EXCEPTION

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1)
      VALUES (id_log.NextVal, 'Transaccio_Mostrar', rsp);
    
END Transaccio_Mostrar;
/

/*****************************************************************************
** Procediment Grup_EsborrarRegal (prod, prov, tit_hom, usu_hom, rsp)       **
******************************************************************************
** Entrada:                                                                 **
**  prod (VARCHAR2): Producte a esborrar de la llista de regals.            **
**  prov (VARCHAR2): Proveidor del producte.                                **
**  tit_hom (VARCHAR2): Grup de regals on es participa.                     **
**  usu_hom (VARCHAR2): Usuari homenatjat del grup.                         **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Esborra un regal d'un grup i en retorna totes les aportacions. En cas    **
** d'estar repetit, n'esborra el ms nou, per afectar als menys usuaris     **
** possibles.                                                               **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE Grup_EsborrarRegal (prod IN VARCHAR2, prov IN VARCHAR2,
  tit_hom IN VARCHAR2, usu_hom IN VARCHAR2, rsp OUT VARCHAR2)
AS

  codi_desig Desig.codi%TYPE;
  tr_tipus Transaccio.tipus%TYPE;
  tr_part Transaccio.participant%TYPE;
  tr_dest Transaccio.dest_regal%TYPE;
  tr_mode Transaccio.mode_p%TYPE;
  tr_dades Transaccio.dades_p%TYPE;
  tr_quant Transaccio.quantitat%TYPE;
  
  --Cursor que recupera els regals que poden ser esborrats d'un grup
  CURSOR regals_esborrables IS
    SELECT d.codi
    FROM Regal r JOIN Desig d
    ON d.codi = r.desig
    WHERE d.producte = prod AND d.proveidor = prov
    AND r.titol_hom = tit_hom AND r.usuari_hom = usu_hom
    AND (r.estat = 'AFEGIT_A_UN_GRUP' OR r.estat = 'PAGAMENT_COMPLETAT')
    ORDER BY d.data_pet desc; 
  
  --Cursor que recupera les transaccions assignades a un regal
  CURSOR transaccions_regal IS
    SELECT t.tipus, t.participant, t.dest_regal, t.mode_p, t.dades_p, t.quantitat
    FROM Transaccio t
    WHERE t.regal = codi_desig
      AND t.titol_hom = tit_hom AND t.usuari_hom = usu_hom;

  REGAL_INCORRECTE EXCEPTION;

BEGIN

  rsp := 'OK';
  --Si un regal est en estat ELIMINAT, NO_SUFICIENTS_FONS, ENVIAT o REBUT no
  --pot ser cancellat, doncs no son estats vlids.
  OPEN regals_esborrables;
  --Per com est ordenat, s el ms nou.
  FETCH regals_esborrables INTO codi_desig;
  CLOSE regals_esborrables;
  IF codi_desig IS NULL THEN RAISE REGAL_INCORRECTE;
  END IF;
  --Marquem el regal com a eliminat
  UPDATE Regal r
    SET r.estat = 'ELIMINAT'
    WHERE r.desig = codi_desig;
  --I el desig associat com a pendent
  UPDATE Desig d
    SET d.estat = 'PENDENT', d.data_estat = SYSDATE
    WHERE d.codi = codi_desig;
  --El sistema retorna totes les transaccions associades,
  OPEN transaccions_regal;
    LOOP
      FETCH transaccions_regal INTO tr_tipus, tr_part, tr_dest, tr_mode, 
        tr_dades, tr_quant;
      EXIT WHEN transaccions_regal%NOTFOUND;
      Transaccio_Afegir ('Devoluci', tr_part, tit_hom, usu_hom, 
        codi_desig, tr_dest, tr_mode, tr_dades, tr_quant, rsp);
    END LOOP;
  CLOSE transaccions_regal;
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4)
    VALUES (id_log.NextVal, 'Grup_EsborrarRegal', rsp, prod, prov, tit_hom, usu_hom);
  
EXCEPTION

  WHEN REGAL_INCORRECTE THEN
    rsp := 'ERROR: El regal o el seu estat no son correctes.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4)
      VALUES (id_log.NextVal, 'Grup_EsborrarRegal', rsp, prod, prov, tit_hom, usu_hom);
    
  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4)
      VALUES (id_log.NextVal, 'Grup_EsborrarRegal', rsp, prod, prov, tit_hom, usu_hom);

END Grup_EsborrarRegal;
/

/*****************************************************************************
** Procediment Grup_ConsultarPendents(rsp)                                  **
******************************************************************************
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Retorna el llistat dels grups amb regals pendents.                       **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE Grup_ConsultarPendents (rsp OUT VARCHAR2)
AS

  titol_actual Grup.titol%TYPE;
  hom_actual Grup.homenatjat%TYPE;
  suma_regals Regal.preu%TYPE;
  suma_aportat Participacio.aportacio%TYPE;
  total_pendent Regal.preu%TYPE;
  perc_aconseguit DECIMAL(5,2);
  --Cursor que recupera un sol cop el nom dels grups que tenen regals pendents
  CURSOR grups_pendents IS
    SELECT UNIQUE g.titol, g.homenatjat
    FROM Regal r JOIN Grup g
    ON r.titol_hom = g.titol AND r.usuari_hom = g.homenatjat
    WHERE r.estat = 'AFEGIT_A_UN_GRUP'
    ORDER BY g.titol asc, g.homenatjat asc;
    
BEGIN

  rsp := 'OK';
  --Mostrem una capalera
  DBMS_OUTPUT.PUT_LINE('Llista de grups actius amb regals pendents');
  DBMS_OUTPUT.PUT_LINE('');
  DBMS_OUTPUT.PUT_LINE(
    RPAD('Nom del grup',40) ||
    RPAD('Homenatjat',25) ||
    RPAD('Total Regals',20) ||
    RPAD('Total Pendent',20) ||
    RPAD('% Aconseguit', 15));
  --I mostrem els grups
  OPEN grups_pendents;
  LOOP
    --Obtenim el primer grup amb regals pendents.
    FETCH grups_pendents INTO titol_actual, hom_actual;
    EXIT WHEN grups_pendents%NOTFOUND;
    --Obtenim l'import acumulat del grup.
    SELECT COALESCE(sum(r.preu),0) INTO suma_regals
    FROM Regal r JOIN Grup g
    ON r.titol_hom = g.titol AND r.usuari_hom = g.homenatjat
    WHERE g.titol = titol_actual AND g.homenatjat = hom_actual;
    --Obtenim la quantitat acumulada pel grup de regals.
    SELECT COALESCE(sum(p.aportacio),0) INTO suma_aportat
    FROM Regal r JOIN Grup g
    ON r.titol_hom = g.titol AND r.usuari_hom = g.homenatjat
    LEFT JOIN Participacio p
    ON p.desig = r.desig AND p.titol_hom = g.titol AND p.usuari_hom = g.homenatjat
    WHERE g.titol = titol_actual AND g.homenatjat = hom_actual;
    --Operem per trobar les dades calculades i mostrem el resultat demanat
    total_pendent := suma_regals - suma_aportat;
    perc_aconseguit := suma_aportat / suma_regals * 100;
    DBMS_OUTPUT.PUT_LINE(
    RPAD(titol_actual,40) ||
    RPAD(hom_actual,24) ||
    to_char(suma_regals, '999G999D99') || RPAD(' ',10) ||
    to_char(total_pendent, '999G999D99') || RPAD(' ',13) ||
    to_char(perc_aconseguit, '990D99') || RPAD('%',8));
  END LOOP;
  CLOSE grups_pendents;
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1)
    VALUES (id_log.NextVal, 'Grup_ConsultarPendents', rsp);

EXCEPTION

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1)
      VALUES (id_log.NextVal, 'Grup_ConsultarPendents', rsp);
    
END;
/

/*****************************************************************************
** Procediment Regal_Enviar(producte, proveidor, rsp)                       **
******************************************************************************
** Entrada:                                                                 **
**  prod (VARCHAR2): Producte a enviar.                                     **
**  prov (VARCHAR2): Proveidor del producte.                                **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Demana a un proveidor que envii un producte concret.                     **
*****************************************************************************/
CREATE OR REPLACE
FUNCTION Regal_Enviar (prod IN VARCHAR2, prov IN VARCHAR2, rsp OUT VARCHAR2)
  RETURN VARCHAR
IS

  answer VARCHAR(100);
  pr Producte.nom%TYPE;
  veg INT;
  
BEGIN

  rsp := 'OK';
  --Implementar aqui la solicitud d'enviament d'un regal segons interessi al
  --client (mitjanant e-mail al proveidor, per exemple). El procediment re-
  --torna OK si el proveidor confirma l'enviament, o ERROR en cas contrari.
  answer := 'OK';
    
  --Actualitzar mdul estadstic. Si un regal s'enva es considera adquirit.
    --Actualitzaci del nombre de vegades que s'ha regalat un producte.
  EST_ActualitzaRegalVegades (prod, rsp);
    --Actualitzaci del nombre total de regals adquirits durant un any.
  EST_ActualitzaNumRegalsAny (rsp);
    --Actualitzaci de l'import acumulat en regals durant un any.
  EST_ActualitzaImportRegalsAny (prod, prov, rsp);
  
  --Retorn de la resposta del proveidor  
  RETURN answer;
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
    VALUES (id_log.NextVal, 'Regal_Enviar', rsp, prod, prov);
  
EXCEPTION

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
      VALUES (id_log.NextVal, 'Regal_Enviar', rsp, prod, prov);
    
END Regal_Enviar;
/

/*****************************************************************************
** Procediment Regal_ConfirmarRebut()                                       **
******************************************************************************
** Entrada:                                                                 **
**  usr (VARCHAR2): Usuari receptor del regal.                              **
**  prod (VARCHAR2): Producte a enviar.                                     **
**  prov (VARCHAR2): Proveidor del producte.                                **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Permet a un usuari receptor indicar al sistema que ha rebut un regal.    **
** El marca com a rebut en la llista de regals, i marca el desig associat   **
** com a aconseguit.                                                        **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE Regal_ConfirmarRebut (usr IN VARCHAR2, prod IN VARCHAR2, 
  prov IN VARCHAR2, rsp OUT VARCHAR2)
AS
  regal_actual Desig.codi%TYPE;
  --Cursor que recupera un regal enviat concret d'un grup
  CURSOR regals_enviats IS
    SELECT r.desig
    FROM Regal r JOIN Desig d
    ON r.desig = d.codi JOIN Grup g
    ON r.titol_hom = g.titol AND r.usuari_hom = g.homenatjat
    WHERE r.estat = 'ENVIAT'
      AND g.receptor = usr
      AND d.producte = prod AND d.proveidor = prov
    ORDER BY d.data_pet asc;
    
  REGAL_INCORRECTE EXCEPTION;
    
BEGIN

  rsp := 'OK';
  --Obtenim el regal coincident pertanyent al desig ms antic.
  OPEN regals_enviats;
    FETCH regals_enviats INTO regal_actual;
  CLOSE regals_enviats;
  IF regal_actual IS NULL THEN RAISE REGAL_INCORRECTE;
  END IF;
  --Canviem l'estat a rebut.
  UPDATE Regal r
  SET r.estat = 'REBUT'
  WHERE r.desig = regal_actual;
  --El desig de l'homenatjat tamb est aconseguit.
  UPDATE Desig d
  SET d.estat = 'ACONSEGUIT', d.data_estat = SYSDATE
  WHERE d.codi = regal_actual;
  
  --Actualitzar mdul estadstic
  EST_ActualitzaUsuariRegals (rsp);

  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3)
    VALUES (id_log.NextVal, 'Regal_ConfirmarRebut', rsp, usr, prod, prov);

EXCEPTION

  WHEN REGAL_INCORRECTE THEN
    rsp := 'ERROR: El regal o el destinatari no son correctes.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3)
      VALUES (id_log.NextVal, 'Regal_ConfirmarRebut', rsp, usr, prod, prov);
    
  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3)
      VALUES (id_log.NextVal, 'Regal_ConfirmarRebut', rsp, usr, prod, prov);
  
END Regal_ConfirmarRebut;
/

/*****************************************************************************
** Procediment Regal_ConsultarRebuts(usr, rsp)                              **
******************************************************************************
** Entrada:                                                                 **
**  usr (VARCHAR2): Usuari receptor de regals.                              **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Mostra la llista de regals rebuts per l'usuari indicat.                  **
** Cal tenir en compte que l'usuari indicat s el receptor dels regals, i   **
** no s necessriament l'usuari al qui estan destinats finalment.          **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE Regal_ConsultarRebuts (usr IN VARCHAR2, rsp OUT VARCHAR2)
AS

  --Cursor que recupera els regals rebuts per un usuari
  CURSOR regals_rebuts IS
    SELECT d.producte, d.proveidor
    FROM Grup g JOIN Regal r
    ON g.titol = r.titol_hom AND g.homenatjat = r.usuari_hom
    JOIN Desig d
    ON r.desig = d.codi
    WHERE r.estat = 'REBUT' AND g.receptor = usr
    ORDER BY d.proveidor desc, d.producte desc;
    
BEGIN

  rsp := 'OK';
  --Mostrem la capalera
  DBMS_OUTPUT.PUT_LINE('Llista regals rebuts per lusuari ' || usr );
  DBMS_OUTPUT.PUT_LINE('');
  DBMS_OUTPUT.PUT_LINE(RPAD('Producte',30)||RPAD('Provedor',25));
  --I mostrem les dades dels regals
  FOR registre IN regals_rebuts LOOP
    DBMS_OUTPUT.PUT_LINE(RPAD(registre.producte,30) || RPAD(registre.proveidor,18)); 
  END LOOP;
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
    VALUES (id_log.NextVal, 'Regal_ConsultarRebuts', rsp, usr);

EXCEPTION
    
  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
      VALUES (id_log.NextVal, 'Regal_ConsultarRebuts', rsp, usr);
    
END Regal_ConsultarRebuts;
/

/*****************************************************************************
** Procediment Desig_ConsultarRebuts(usr, rsp)                              **
******************************************************************************
** Entrada:                                                                 **
**  usr (VARCHAR2): Usuari del sistema.                                     **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Mostra la llista de regals aconseguits per l'usuari final indicat.       **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE Desig_ConsultarRebuts (usr IN VARCHAR2, rsp OUT VARCHAR2)
AS

  --Cursor que recupera els desitjos aconseguits per un usuari (regals)
  CURSOR regals_rebuts IS
    SELECT d.producte, d.proveidor
    FROM Desig d
    WHERE d.estat = 'ACONSEGUIT' AND d.usuari = usr
    ORDER BY d.proveidor desc, d.producte desc;
    
BEGIN

  rsp := 'OK';
  --Mostrem una capalera
  DBMS_OUTPUT.PUT_LINE('Llista regals aconseguits per lusuari ' || usr );
  DBMS_OUTPUT.PUT_LINE('');
  DBMS_OUTPUT.PUT_LINE(RPAD('Producte',30)||RPAD('Provedor',25));
  --I mostrem la llista de regals de l'usuari
  FOR registre IN regals_rebuts LOOP
    DBMS_OUTPUT.PUT_LINE(RPAD(registre.producte,30) ||
    RPAD(registre.proveidor,18)); 
  END LOOP;
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
    VALUES (id_log.NextVal, 'Desig_ConsultarRebuts', rsp, usr);

EXCEPTION

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
      VALUES (id_log.NextVal, 'Desig_ConsultarRebuts', rsp, usr);
    
END Desig_ConsultarRebuts;
/

/*****************************************************************************
** Procediment Participacio_Aportar(prod, prov, homenatge, homenatjat,      **
**    usuari, aport, mode_pag, dades_pag, rsp)                              **
******************************************************************************
** Entrada:                                                                 **
**  prod (VARCHAR2): Producte en el que es participa.                       **
**  prov (VARCHAR2): Proveidor del producte.                                **
**  homenatge (VARCHAR2): Grup de regals on es participa.                   **
**  homenatjat (VARCHAR2): Usuari homenatjat del grup.                      **
**  usuari (VARCHAR2): usuari que fa la aportaci.                          **
**  aport (DECIMAL): Import amb que es participa en el regal.               **
**  mode_pag (VARCHAR2): Forma de pagament.                                 **
**  dades_pag (VARCHAR2): Info del mode de pagament (compte PayPal, etc).   **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Permet que un usuari d'un grup pugui participar en un regal. Si ja ha    **
** participat en el regal indicat, en corregeix la aportaci. Per tant,     **
** noms es mostra una aportaci amb l'import total. Per altra banda, es    **
** desen tantes transaccions com vegades es participi.                      **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE Participacio_Aportar (prod IN VARCHAR2, prov IN VARCHAR2, 
  homenatge IN VARCHAR2, homenatjat IN VARCHAR2, usuari IN VARCHAR2,
  aport IN DECIMAL, mode_pag IN VARCHAR2, dades_pag IN VARCHAR2,
  rsp OUT VARCHAR2)
  
AS

  aport_final Participacio.aportacio%TYPE;
  regal_actual Regal.desig%TYPE;
  import_acumulat Participacio.aportacio%TYPE;
  preu_regal Regal.preu%TYPE;
  dest_regal Grup.receptor%TYPE;
  existeix INT;
  inactiu Grup.finalitzat%TYPE;
  
  INSERT_NULL_VALUES EXCEPTION;
  IMPORT_NEGATIU EXCEPTION;
  REGAL_INCORRECTE EXCEPTION;
  GRUP_INACTIU EXCEPTION;
  USUARI_INCORRECTE EXCEPTION;
  
  PRAGMA EXCEPTION_INIT(INSERT_NULL_VALUES, -1400);
  PRAGMA EXCEPTION_INIT(IMPORT_NEGATIU, -2290);  
  
  --Cursor que ens retorna la llista de regals pendents i la quantitat aportada.
  --El primer registre ser el que tingui ms import aconseguit, i ser el que
  --es modificar. Si no hi ha registres, o b no hi ha regals, o estn aconse-
  --guits s'ha de llenar excepci, doncs la participaci no s vlida.
  CURSOR regals_pendents IS
    SELECT r.desig, r.preu, COALESCE(sum(p.aportacio),0) AS Acumulat
    --Primer join, relaciona el codi del regal amb el producte associat.
    FROM Regal r JOIN Desig d
    ON r.desig = d.codi
    --No tots els regals necessriament han de tenir ja una participaci.
    LEFT JOIN Participacio p
    ON r.desig = p.desig
    --L'nic estat vlid per a aconseguir un regal s 'AFEGIT_A_UN_GRUP'.
    WHERE r.estat='AFEGIT_A_UN_GRUP'
      --Verifiquem que el regal sigui de l'homenatge correcte. En cas contrari,
      --es podria participar en grups dels quals no es participa.
      AND r.titol_hom = homenatge
      AND r.usuari_hom = homenatjat
      --Seleccionem nicament el producte desitjat
      AND d.producte = prod
      AND d.proveidor = prov
    GROUP BY r.desig, r.preu, r.titol_hom, r.usuari_hom, d.producte, d.proveidor, d.data_pet
    --Ens assegurem que la primera fila sigui el regal amb ms aportaci.
    --Si no hi han aportacions, ser el desig ms antic.
    ORDER BY Acumulat desc, d.data_pet asc;
    
BEGIN

  --Verifiquem que el grup no estigui de baixa
  SELECT finalitzat INTO inactiu
  FROM Grup g
  WHERE g.titol = homenatge AND g.homenatjat = homenatjat;
  IF inactiu = 'T' THEN RAISE GRUP_INACTIU;
  END IF;

  --Verifiquem si l'usuari indicat pertany al grup
  SELECT COUNT (*) INTO existeix
  FROM Integrant i
  WHERE i.titol_grup = homenatge AND i.hom_grup = homenatjat 
    AND i.usu_grup = usuari;
  IF existeix = 0 THEN RAISE USUARI_INCORRECTE;
  END IF;
  
  rsp := 'OK';
  OPEN regals_pendents;
  FETCH regals_pendents INTO regal_actual, preu_regal, import_acumulat;
  CLOSE regals_pendents;
  IF regal_actual IS NULL THEN RAISE REGAL_INCORRECTE;
  END IF;
  --Ara sabem el regal en que hem de participar, i el que falta per aconseguir-lo.
  --Si l'usuari aporta ms del necessari, el sistema ho ajusta al mxim.
  IF aport > preu_regal - import_acumulat
    THEN aport_final := preu_regal - import_acumulat;
  ELSE aport_final := aport;
  END IF;
  --Ara que la aportaci s correcta, hem de mirar si l'usuari actual ja ha
  --participat anteriorment en el regal o no, i actuar en conseqncia.
  --Si ha participat, el tractament d'excepcions s'encarrega de l'UPDATE.
  SELECT receptor INTO dest_regal
  FROM Grup g
  WHERE g.titol = homenatge
    AND g.homenatjat =  homenatjat;
  INSERT INTO Participacio (desig, titol_hom, usuari_hom, participant, aportacio)
  VALUES (regal_actual, homenatge, homenatjat, usuari, aport_final);
  --Finalment, desem la transacci. Necessitem trobar el destinatari.
  Transaccio_Afegir ('Pagament', usuari, homenatge, homenatjat,
    regal_actual, dest_regal, mode_pag, dades_pag, aport_final, rsp);
  IF import_acumulat + aport_final = preu_regal
    THEN UPDATE Regal r SET r.estat = 'PAGAMENT_COMPLETAT'
      WHERE r.desig = regal_actual;
  END IF;
  
  --Actualitzar mdul estadstic
    --Actualitzem l'usuari amb ms participacions
  EST_ActualitzaUsrMaxPart (rsp);
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4,
    ent5, ent6, ent7, ent8)
    VALUES (id_log.NextVal, 'Participacio_Aportar', rsp, prod, prov, homenatge,
      homenatjat, usuari, aport, mode_pag, dades_pag);
   
EXCEPTION

  --Excepci que s'encarrega de corregir l'import participa total per un usuari
  WHEN DUP_VAL_ON_INDEX THEN
    UPDATE Participacio p
    SET p.aportacio = aportacio + aport_final
    WHERE p.desig = regal_actual
      AND p.titol_hom = homenatge
      AND p.usuari_hom = homenatjat
      AND p.participant = usuari;
    Transaccio_Afegir ('Pagament', usuari, homenatge, homenatjat,
      regal_actual, dest_regal, mode_pag, dades_pag, aport_final, rsp);
    IF import_acumulat + aport_final = preu_regal
      THEN UPDATE Regal r SET r.estat = 'PAGAMENT_COMPLETAT'
        WHERE r.desig = regal_actual;
    END IF;
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4,
      ent5, ent6, ent7, ent8)
      VALUES (id_log.NextVal, 'Participacio_Aportar', rsp, prod, prov, homenatge,
        homenatjat, usuari, aport, mode_pag, dades_pag);
  
  WHEN INSERT_NULL_VALUES THEN
    rsp := 'ERROR: El regal no existeix o ja est aconseguit.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4,
      ent5, ent6, ent7, ent8)
      VALUES (id_log.NextVal, 'Participacio_Aportar', rsp, prod, prov, homenatge,
        homenatjat, usuari, aport, mode_pag, dades_pag);
        
  WHEN NO_DATA_FOUND THEN
    rsp := 'ERROR: El grup indicat s incorrecte o no existeix.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4,
      ent5, ent6, ent7, ent8)
      VALUES (id_log.NextVal, 'Participacio_Aportar', rsp, prod, prov, homenatge,
        homenatjat, usuari, aport, mode_pag, dades_pag);
    
  WHEN IMPORT_NEGATIU THEN
    rsp := 'ERROR: Una aportaci no pot ser negativa.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4,
      ent5, ent6, ent7, ent8)
      VALUES (id_log.NextVal, 'Participacio_Aportar', rsp, prod, prov, homenatge,
        homenatjat, usuari, aport, mode_pag, dades_pag);
    
  WHEN REGAL_INCORRECTE THEN
    rsp := 'ERROR: El regal no existeix o ja est aconseguit.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4,
      ent5, ent6, ent7, ent8)
      VALUES (id_log.NextVal, 'Participacio_Aportar', rsp, prod, prov, homenatge,
        homenatjat, usuari, aport, mode_pag, dades_pag);
    
  WHEN GRUP_INACTIU THEN
    rsp := 'ERROR: No es pot participar en un grup donat de baixa.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4,
      ent5, ent6, ent7, ent8)
      VALUES (id_log.NextVal, 'Participacio_Aportar', rsp, prod, prov, homenatge,
        homenatjat, usuari, aport, mode_pag, dades_pag);
  
  WHEN USUARI_INCORRECTE THEN
    rsp := 'ERROR: Usuari incorrecte o no pertanyent al grup.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4,
      ent5, ent6, ent7, ent8)
      VALUES (id_log.NextVal, 'Participacio_Aportar', rsp, prod, prov, homenatge,
        homenatjat, usuari, aport, mode_pag, dades_pag);
  
  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2, ent3, ent4,
      ent5, ent6, ent7, ent8)
      VALUES (id_log.NextVal, 'Participacio_Aportar', rsp, prod, prov, homenatge,
        homenatjat, usuari, aport, mode_pag, dades_pag);

END Participacio_Aportar;
/

/*****************************************************************************
** Procediment Participacio_MillorsUsuaris (rsp)                            **
******************************************************************************
** Entrada:                                                                 **
**  El procediment no necessita parmetres d'entrada.                       **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Mostra una llista amb els usuaris amb participacions en regals per sobre **
** de la mitja. El rnking no t en compte les devolucions per regals, ja   **
** que es t en consideraci la intenci de participar en el regal.         **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE Participacio_MillorsUsuaris (rsp OUT VARCHAR2)
AS

  mitja Participacio.aportacio%TYPE;
  usuari Participacio.participant%TYPE;
  aportat Participacio.aportacio%TYPE;
  ranking DECIMAL (10,0) := 0;
  --Cursor que recupera els participants i aportacions ordenats de major a menor
  CURSOR participants_i_aportacions IS
    SELECT p.participant, sum(p.aportacio) AS Aportat
    FROM participacio p
    GROUP BY p.participant
    ORDER BY aportat desc;
    
BEGIN

  rsp := 'OK';
  --Primer trobem la mitja d'aportacions
  SELECT avg(Aportat) INTO mitja FROM
  (SELECT p.participant, sum(p.aportacio) AS Aportat
  FROM participacio p
  GROUP BY p.participant
  ORDER BY Aportat desc);
  --Mostrem una capalera
  DBMS_OUTPUT.PUT_LINE('');
  DBMS_OUTPUT.PUT_LINE('Llista dusuaris amb participacions per sobre de la mitja (' || mitja || ')');
  DBMS_OUTPUT.PUT_LINE('');
  DBMS_OUTPUT.PUT_LINE(
    RPAD('Posici', 12) ||
    RPAD('Total Aportat',20) ||
    RPAD('Usuari',35));
  --Ara podem comparar cada usuari, i mostrar-lo si s'escau.
  OPEN participants_i_aportacions;
  LOOP
    FETCH participants_i_aportacions INTO usuari, aportat;
    EXIT WHEN participants_i_aportacions%NOTFOUND;
    IF aportat > mitja THEN
      ranking := ranking + 1;
      DBMS_OUTPUT.PUT_LINE(
        RPAD(ranking, 12) ||
        to_char (aportat, '999G999D99') || RPAD(' ',9) ||
        RPAD(usuari,35)); 
    END IF;
  END LOOP;  
  CLOSE participants_i_aportacions;
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1)
    VALUES (id_log.NextVal, 'Participacio_MillorsUsuaris', rsp);

EXCEPTION

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1)
      VALUES (id_log.NextVal, 'Participacio_MillorsUsuaris', rsp);

END;
/

/*****************************************************************************
** Procediment Grup_ConsultarUrgents (rsp)                                  **
******************************************************************************
** Entrada:                                                                 **
**  El procediment no necessita parmetres d'entrada.                       **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Retorna una llista amb els grups en data limit de recopilaci de fons.   **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE Grup_ConsultarUrgents (rsp OUT VARCHAR2)
AS

  tit_grup Grup.titol%TYPE;
  hom_grup Grup.homenatjat%TYPE;
  --Cursor que recupera els grups que es troben en data limit de regals
  CURSOR urgents IS
    SELECT g.titol, g.homenatjat
    FROM Grup g
    WHERE g.data_fons <= SYSDATE;
    
BEGIN

  rsp := 'OK';
  --Mostrem una capalera
  DBMS_OUTPUT.PUT_LINE('Llista de grups en data lmit de fons (' || SYSDATE || ')');
  DBMS_OUTPUT.PUT_LINE('');
  DBMS_OUTPUT.PUT_LINE(RPAD('Nom del grup',40) || RPAD('Homenatjat',25));
  --I mostrem els grups
  OPEN urgents;
  LOOP
    FETCH urgents INTO tit_grup, hom_grup;
    EXIT WHEN urgents%NOTFOUND;
    DBMS_OUTPUT.PUT_LINE(RPAD(tit_grup,40) || RPAD(hom_grup,24));
  END LOOP;
  CLOSE urgents;
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1)
    VALUES (id_log.NextVal, 'Grup_ConsultarUrgents', rsp);
  
EXCEPTION

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1)
      VALUES (id_log.NextVal, 'Grup_ConsultarUrgents', rsp);
    
END Grup_ConsultarUrgents;
/

/*****************************************************************************
** Procediment Grup_ForcarBaixa (tit_homenatge, usu_homenatge, rsp)         **
******************************************************************************
** Entrada:                                                                 **
**  tit_homenatge (VARCHAR2): Nom del grup de regals.                       **
**  usu_homenatge (VARCHAR2): Usuari homenatjat pel grup.                   **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Procediment utilitzat pel sistema per forar la baixa d'un grup, com     **
** quan ha exhaurit el temps per a paticipar en els regals. Envia els       **
** regals aconseguits i retorna els imports dels no aconseguits.            **
*****************************************************************************/
CREATE OR REPLACE PROCEDURE Grup_ForcarBaixa(tit_homenatge IN VARCHAR2,
  usu_homenatge IN VARCHAR2, rsp OUT VARCHAR2)
AS

  resultat_enviament VARCHAR(100);
  regal_actual Regal.desig%TYPE;
  producte_act Desig.producte%TYPE;
  proveidor_act Desig.proveidor%TYPE;
  tr_tipus Transaccio.tipus%TYPE;
  tr_part Transaccio.participant%TYPE;
  tr_dest Transaccio.dest_regal%TYPE;
  tr_mode Transaccio.mode_p%TYPE;
  tr_dades Transaccio.dades_p%TYPE;
  tr_quant Transaccio.quantitat%TYPE;
  estat_grup Grup.finalitzat%TYPE;
  aconseguits Regal_any.adquirits%TYPE; --Necessari pel mdul estadstic
  fallits Regal_any.adquirits%TYPE;     --Necessari pel mdul estadstic

  GRUP_FINALITZAT EXCEPTION;

  --Cursor per recuperar els regals assolits per el grup indicat.
  CURSOR regals_aconseguits IS
    SELECT r.desig, d.producte, d.proveidor
    FROM Regal r JOIN Desig d ON r.desig = d.codi
    WHERE r.titol_hom = tit_homenatge AND r.usuari_hom = usu_homenatge
      AND r.estat = 'PAGAMENT_COMPLETAT';
  
  --Cursor per obtenir els regals que no han estat assolits pel grup.
  CURSOR regals_fallits IS
    SELECT r.desig
    FROM Regal r
    WHERE r.titol_hom = tit_homenatge AND r.usuari_hom = usu_homenatge
      AND r.estat = 'AFEGIT_A_UN_GRUP';
  
  --Cursor que obt les transaccions associades a un regal.
  CURSOR transaccions_regal IS
    SELECT t.tipus, t.participant, t.dest_regal, t.mode_p, t.dades_p, t.quantitat
    FROM Transaccio t
    WHERE t.regal = regal_actual
      AND t.titol_hom = tit_homenatge AND t.usuari_hom = usu_homenatge;

BEGIN

  rsp := 'OK';
  
  SELECT finalitzat INTO estat_grup
  FROM Grup
  WHERE titol = tit_homenatge AND homenatjat = usu_homenatge;
  
  IF estat_grup = 'T' THEN RAISE GRUP_FINALITZAT;
  END IF;
  
  aconseguits := 0;
  fallits := 0;
  --El sistema realitza les devolucions pertinents i desassigna els desitjos.
  OPEN regals_fallits;
  LOOP
    FETCH regals_fallits INTO regal_actual;
    EXIT WHEN regals_fallits%NOTFOUND;
    fallits := fallits + 1;  --Incrementem el nombre de regals no aconseguits
    --Desassigna el desig
    UPDATE Desig d
    SET d.estat = 'PENDENT', d.data_estat = SYSDATE
    WHERE d.codi = regal_actual;
    --I retorna totes les transaccions;
    OPEN transaccions_regal;
    LOOP
      FETCH transaccions_regal INTO tr_tipus, tr_part, tr_dest, tr_mode, 
        tr_dades, tr_quant;
      EXIT WHEN transaccions_regal%NOTFOUND;
      Transaccio_Afegir ('Devoluci', tr_part, tit_homenatge, usu_homenatge, 
        regal_actual, tr_dest, tr_mode, tr_dades, tr_quant, rsp);
    END LOOP;
    CLOSE transaccions_regal;
  END LOOP;
  CLOSE regals_fallits;
  
  --Es marquen els regals no aconseguits com a tal
  UPDATE Regal r
  SET r.estat = 'NO_SUFICIENTS_FONS'
  WHERE r.titol_hom = tit_homenatge AND r.usuari_hom = usu_homenatge
    AND r.estat = 'AFEGIT_A_UN_GRUP';

  --Desprs el sistema gestiona l'enviament dels regals.
  OPEN regals_aconseguits;
  LOOP
    FETCH regals_aconseguits INTO regal_actual, producte_act, proveidor_act;
    EXIT WHEN regals_aconseguits%NOTFOUND;
    aconseguits := aconseguits + 1;  --Incrementem el nombre d'aconseguits
    resultat_enviament := Regal_Enviar(producte_act, proveidor_act, rsp);
    IF resultat_enviament = 'OK' THEN
      UPDATE Regal r
      SET r.estat = 'ENVIAT'
      WHERE r.desig = regal_actual;
    END IF;
  END LOOP;
  CLOSE regals_aconseguits;
  
  --Actualitzaci del mdul estadstic.
    --Actualitzaci del percentatge entre regals aconseguits i fallits
  EST_ActualitzaAconseguitsAny (aconseguits, fallits, rsp);
  
  --Finalment, es marca el grup com a finalitzat
  UPDATE Grup g
  SET g.finalitzat = 'T'
  WHERE g.titol = tit_homenatge AND g.homenatjat = usu_homenatge;
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
    VALUES (id_log.NextVal, 'Grup_ForcarBaixa', rsp, tit_homenatge, usu_homenatge);
  
EXCEPTION

  WHEN GRUP_FINALITZAT THEN
    rsp := 'ERROR: El grup ja est donat de baixa.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
      VALUES (id_log.NextVal, 'Grup_ForcarBaixa', rsp, tit_homenatge, usu_homenatge);
      
  WHEN NO_DATA_FOUND THEN
    rsp := 'ERROR: El grup indicat no existeix.';
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
      VALUES (id_log.NextVal, 'Grup_ForcarBaixa', rsp, tit_homenatge, usu_homenatge);

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
      VALUES (id_log.NextVal, 'Grup_ForcarBaixa', rsp, tit_homenatge, usu_homenatge);
    
END Grup_ForcarBaixa;
/

/*****************************************************************************
** Procediment Grup_GestionarUrgents                                        **
******************************************************************************
** Entrada:                                                                 **
**  tit_homenatge (VARCHAR2): Nom del grup de regals.                       **
**  usu_homenatge (VARCHAR2): Usuari homenatjat pel grup.                   **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Procediment utilitzat pel sistema a diari per a gestionar els grups de   **
** regals. Recorre tots els grups en data fi de fons, i gestiona els regals **
** de cadascun mitjanant el procediment Grup_ForcarBaixa().                **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE Grup_GestionarUrgents (rsp OUT VARCHAR2)
AS

  tit_grup Grup.titol%TYPE;
  hom_grup Grup.homenatjat%TYPE;
  
  --Cursor que recupera els grups en data lmit de fons
  CURSOR urgents IS
    SELECT g.titol, g.homenatjat
    FROM Grup g
    WHERE g.data_fons <= SYSDATE;
    
BEGIN

  rsp := 'OK';
  --Recuperem els grups
  OPEN urgents;
  LOOP
    FETCH urgents INTO tit_grup, hom_grup;
    EXIT WHEN urgents%NOTFOUND;
    --Forcem la bai(i la gesti de regals) de cada grup en data lmit
    Grup_ForcarBaixa (tit_grup, hom_grup, rsp);
  END LOOP;
  CLOSE urgents;
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1)
    VALUES (id_log.NextVal, 'Grup_GestionarUrgents', rsp);
  
EXCEPTION

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1)
      VALUES (id_log.NextVal, 'Grup_GestionarUrgents', rsp);
    
END Grup_GestionarUrgents;
/

/*****************************************************************************
** Procediment Llistar_Ciutats (d_any, rsp)                                 **
******************************************************************************
** Entrada:                                                                 **
**  d_any (INT): Any que es vol consultar                                   **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Mostra una llista de les 10 ciutats on ms regals s'han rebut durant     **
** l'any sollicitat.                                                       **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE Llistar_Ciutats (d_any INT, rsp OUT VARCHAR2)
AS

  pobl cp_id.ciutat%TYPE;
  acumulat Regal.preu%TYPE;
  num_reg INT;
  iter INT;

  --Cursor que recupera les ciutats amb ms regals rebuts en un any indicat
  CURSOR resultats IS
    SELECT cp.ciutat, SUM(r.preu) AS Valor, COUNT(r.preu)
    FROM Desig d JOIN Regal r
      ON d.codi = r.desig
      JOIN Grup g
        ON g.titol = r.titol_hom AND g.homenatjat = r.usuari_hom
        JOIN cp_id cp
          ON g.cp_id = cp.id
    WHERE EXTRACT(year from d.data_estat) = d_any AND r.estat = 'REBUT'
  GROUP BY cp.ciutat
  ORDER BY Valor desc;
  
BEGIN

  rsp := 'OK';
  --Controlem el nombre d'iteracions, perqu volem noms 10 ciutats
  iter := 1;
  
  --Recuperem la llista de ciutats
  DBMS_OUTPUT.PUT_LINE('Llista de les 10 ciutats amb ms regals rebuts en l''any ' || d_any || '.');
  DBMS_OUTPUT.PUT_LINE('');
  OPEN resultats;
  LOOP
    --Cerquem cada ciutat
    FETCH resultats INTO pobl, acumulat, num_reg;
    --Controlem el nombre de ciutats mostrades
    EXIT WHEN (resultats%NOTFOUND OR iter > 10);
    --I en mostrem les dades.
    DBMS_OUTPUT.PUT_LINE(LPAD(iter,2) || ' - ' || RPAD(pobl,30) || ' - ' || LPAD(acumulat,10) || '' || ' - (' || num_reg || ' regals).');
    iter := iter + 1;
  END LOOP;
  CLOSE resultats;
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
    VALUES (id_log.NextVal, 'Llistar_Ciutats', rsp, d_any);

EXCEPTION

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1)
      VALUES (id_log.NextVal, 'Llistar_Ciutats', rsp, d_any);
  
END Llistar_Ciutats;
/

/*****************************************************************************
** Procediment Consulta_Log (data_inici, data_fi, rsp)                      **
******************************************************************************
** Entrada:                                                                 **
**  data_inici (DATE): Data d'inici per mostrar la taula de log.            **
**  data_fi (DATE): Data final per mostrar la taula de log.                 **
** Sortida:                                                                 **
**  rsp (VARCHAR2): Indicador de resultat d'execuci de procediment.        **
******************************************************************************
** Mostra les entrades de la taula de log entre les dues dates indicades.   **
*****************************************************************************/
CREATE OR REPLACE
PROCEDURE Consulta_Log (data_inici DATE, data_fi DATE, rsp OUT VARCHAR2)
AS
  id_taula NUMBER;
  data_log DATE;
  procediment VARCHAR2(50);
  sor1 VARCHAR2(150);
  ent_text VARCHAR2(200);
  ent1 VARCHAR2(50);
  ent2 VARCHAR2(50);
  ent3 VARCHAR2(50);
  ent4 VARCHAR2(50);
  ent5 VARCHAR2(50);
  ent6 VARCHAR2(50);
  ent7 VARCHAR2(50);
  ent8 VARCHAR2(50);
  ent9 VARCHAR2(50);

  --Cursor que recupera les entrades de la taula
  CURSOR resultats IS
    SELECT *
    FROM Log_proc
    WHERE data_log >= data_inici AND data_log <= data_fi
    ORDER BY id_taula;
    
BEGIN

  rsp := 'OK';
  --Controlem el nombre d'iteracions, perqu volem noms 10 ciutats
  
  --Mostrem una capalera
  DBMS_OUTPUT.PUT_LINE('');
  DBMS_OUTPUT.PUT_LINE('Log entre les dates ' || data_inici || ' i ' || data_fi);
  DBMS_OUTPUT.PUT_LINE('Id - data | Procediment, rsp, parmetres d''entrada');
  DBMS_OUTPUT.PUT_LINE('');
  --Finalment, mostrem les entrades demanades de la taula
    
  OPEN resultats;
  LOOP
    FETCH resultats INTO id_taula, data_log, procediment, sor1, ent_text, ent1,
    ent2, ent3, ent4, ent5, ent6, ent7, ent8, ent9;
    EXIT WHEN resultats%NOTFOUND;
    
    DBMS_OUTPUT.PUT_LINE(id_taula || ' - ' || to_char(data_log,'DD-MM-YYYY HH24:MI:SS') ||
      ' | ' || procediment || 
      ', ' || sor1 ||
      ', ' || COALESCE (ent1, 'n/a') ||
      ', ' || COALESCE (ent2, 'n/a') ||
      ', ' || COALESCE (ent3, 'n/a') ||
      ', ' || COALESCE (ent4, 'n/a') ||
      ', ' || COALESCE (ent5, 'n/a') ||
      ', ' || COALESCE (ent6, 'n/a') ||
      ', ' || COALESCE (ent7, 'n/a') ||
      ', ' || COALESCE (ent8, 'n/a') ||
      ', ' || COALESCE (ent9, 'n/a') ||
      ', ' || COALESCE (ent_text, 'n/a'));
  END LOOP;
  CLOSE resultats;
  
  --Log
  INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
    VALUES (id_log.NextVal, 'Consulta_Log', rsp, data_inici, data_fi);

EXCEPTION

  WHEN OTHERS THEN
    rsp := 'ERROR: ' || SQLERRM;
    DBMS_OUTPUT.PUT_LINE(rsp);
    --Log
    INSERT INTO Log_proc(id_taula, procediment, sor1, ent1, ent2)
      VALUES (id_log.NextVal, 'Consulta_Log', rsp, data_inici, data_fi);
  
END Consulta_Log;
/